from openai import OpenAI, Stream
from django.conf import settings
from langchain_core.messages import (
    HumanMessage, 
    SystemMessage, 
    ToolMessage, 
    BaseMessage,
    AIMessage,
    messages_from_dict,
    messages_to_dict
    )
from langchain_core.chat_history import InMemoryChatMessageHistory
from langgraph.checkpoint.memory import MemorySaver
from langchain_openai import ChatOpenAI
from langgraph.prebuilt import create_react_agent
from langgraph.graph.graph import CompiledGraph
from django.core.cache import cache
from langchain_core.tools import StructuredTool
from bloomerp.langchain_tools import BaseBloomerpTool
from pydantic import BaseModel, Field


class BloomerpOpenAI:
    def __init__(self, api_key: str = None):
        self.client = OpenAI(api_key=api_key)
        self.model = "gpt-4o-mini"
    
    def _prepare_messages(
            self,
            query: str,
            system_content: str,
            conversation_history: list[dict[str, str]] = None
            ) -> list[dict[str, str]]:
        '''Prepare messages for the OpenAI API.'''
        messages = [
            {'role': 'system', 'content': system_content}
        ]

        if conversation_history:
            messages += conversation_history

        messages.append({'role': 'user', 'content': query})

        return messages

    def is_valid_key(self):
        '''Checks if the OpenAI API key is valid.'''
        try:
            self.client.models.list()
            return True
        except Exception:
            return False

    def create_tiny_mce_content(
            self, 
            prompt:str, 
            stream_response:bool=False,
            conversation_history: list[dict[str, str]] = None
            ) -> str:
        '''Function to create the content for TinyMCE editor.'''

        SYSTEM_CONTENT = '''
        You are a helpful assistant that helps me to create content for TinyMCE editor based on the prompt it gives.
        The output should be the content that can be used in TinyMCE editor, meaning it should be formatted as HTML.
        '''

        # Prepare messages for the OpenAI API
        messages = self._prepare_messages(prompt, SYSTEM_CONTENT, conversation_history)
        
        if stream_response:
            response : Stream = self.client.chat.completions.create(
                model = self.model,
                stream=True,
                messages = messages
            )

            for content in response:
                chunk = content.choices[0].delta.content

                # Filter out the ```html ... ``` from the output
                if chunk:
                    yield chunk
                else:
                    print('Streaming done')
                
        else:
            response = self.client.chat.completions.create(
                model = self.model,
                messages = messages
            )

            content = response.choices[0].message.content
            return content

    def create_sql_query(self,
                        query: str,
                        db_tables_and_columns: list[tuple[str, list[str]]],
                        conversation_history: list[dict[str, str]] = None
                        ) -> str:
        '''
        This function creates a SQL query from a natural language query.
        '''

        SYSTEM_CONTENT = '''
        You are a helpful assistant that helps me to create SQL queries from natural language.
        The output should be a SQL query using sqlite3 syntax, without any explanation. Dont include in the output ```sql ... ```, just the raw SQL query.
        Here are the database tables, columns and datatypes for each column in the database:
        '''

        for table in db_tables_and_columns:
            SYSTEM_CONTENT += f'\n{table[0]}: '

            for column in table[1]:
                SYSTEM_CONTENT += f'{column[0]} ({column[1]}), '
        
        # Prepare messages for the OpenAI API
        messages = self._prepare_messages(query, SYSTEM_CONTENT, conversation_history)

        response = self.client.chat.completions.create(
            model = self.model,
            messages = messages
        )

        sql_query = response.choices[0].message.content

        # Remove ```sql ... ``` from the output
        sql_query = sql_query.replace('```sql\n', '').replace('```', '')

        return sql_query

    def create_document_template(
            self, 
            query: str,
            variables: list[(str, str, str)], 
            conversation_history: list[dict[str, str]] = None,
            stream_response:bool=False
            ) -> str:
        '''
        This function creates a document template from a natural language query.
        '''
        SYSTEM_CONTENT = '''
        You are a helpful assistant that helps me to create document templates from natural language.
        The output should be the document (in html) using jinja2 (Django template) syntax, without any explanation. Dont include in the output ```html ... ```, just the raw HTML.
        For object variables, use the following syntax: {{ object.variable_name }}
        For free variables, use the following syntax: {{ free_variable_name }}

        Here are the available variables that you can use in the document template:
        '''
        for variable in variables:
            SYSTEM_CONTENT += f'\n{variable[0]} ({variable[1]}) - {variable[2]}'


        messages = self._prepare_messages(query, SYSTEM_CONTENT, conversation_history)

        if stream_response:
            response : Stream = self.client.chat.completions.create(
                model = self.model,
                stream=True,
                messages = messages
            )

            for content in response:
                chunk = content.choices[0].delta.content

                # Filter out the ```html ... ``` from the output
                if chunk:
                    yield chunk
        else:
            response = self.client.chat.completions.create(
                model = self.model,
                messages = messages
            )

            content = response.choices[0].message.content
            return content
    
    def create_bloom_ai(self, query: str, conversation_history: list[dict[str, str]] = None) -> str:
        '''
        This function creates a document template from a natural language query.
        '''
        raise NotImplementedError('Bloom AI is not implemented yet')
    
    def create_code(self, query: str, conversation_history: list[dict[str, str]] = None) -> str:
        '''
        This function creates a document template from a natural language query.
        '''

        # Create system content
        SYSTEM_CONTENT = '''
        You are a helpful assistant that helps me to create code from natural language.
        Only return the actual code snippet without giving any explanation.
        Don't include any markdown code blocks in the output.
        '''

        # Prepare messages for the OpenAI API
        messages = self._prepare_messages(query, SYSTEM_CONTENT, conversation_history)

        response = self.client.chat.completions.create(
            model = self.model,
            messages = messages
        )

        code = response.choices[0].message.content

        return code
    


class BloomerpLangChain:
    '''Class to invoke the Bloom AI agent.''' 
    conversation_history : list
    invoked: bool = False

    def __init__(self, 
                 api_key:str=None, 
                 conversation_history:list = None, 
                 ):
        '''
        Args:
            api_key: The OpenAI API key.
            conversation_history: The conversation history.
        '''

        self.api_key = api_key
        self.conversation_history = conversation_history
        self.model = ChatOpenAI(model="gpt-4o-mini", api_key=api_key)

    def invoke_tiny_mce_content(self, query:str):
        '''Invoke the TinyMCE content agent.'''

        SYSTEM_CONTENT = '''You are a helpful assistant that helps me to create content for TinyMCE editor based on the prompt it gives.
        The output should be the content that can be used in TinyMCE editor, meaning it should be formatted as HTML.
        '''

        # Prepare messages for the OpenAI API
        messages = self._prepare_messages(query, SYSTEM_CONTENT)

        # Create the agent
        memory = MemorySaver()
        agent = create_react_agent(self.model, tools=[], checkpointer=memory)

        # Invoke the model
        for chunk in self._invoke_model(messages, agent):
            yield chunk
    
    def invoke_sql_query(self, query:str, db_tables_and_columns:list[tuple[str, list[str]]]):
        '''Invoke the SQL query agent.'''

        # Prepare system content
        SYSTEM_CONTENT = '''You are a helpful assistant that helps me to create SQL queries from natural language.
        The output should be a SQL query using sqlite3 syntax, without any explanation. Dont include in the output ```sql ... ```, just the raw SQL query.
        Here are the database tables, columns and datatypes for each column in the database:
        '''

        for table in db_tables_and_columns:
            SYSTEM_CONTENT += f'\n{table[0]}: '

            for column in table[1]:
                SYSTEM_CONTENT += f'{column[0]} ({column[1]}), '

        # Prepare messages for the OpenAI API
        messages = self._prepare_messages(query, SYSTEM_CONTENT)

        # Create the agent
        memory = MemorySaver()
        agent = create_react_agent(self.model, tools=[], checkpointer=memory)

        # Invoke the model
        for chunk in self._invoke_model(messages, agent):
            yield chunk
        
    def invoke_bloom_ai(self, query:str, tools:list[StructuredTool], user=None):
        '''Invoke the Bloom AI agent.'''

        if user:
            name = user.get_full_name()

        SYSTEM_CONTENT = f'''You are an AI agent for a ERP system called Bloomerp. You are currently serving a user named {name}.
        Your job is to help users with their day-to-day job, which can include answering to queries. 
        You can use the tools provided to you to get information from the database and provide it to the user.
        If you need to get information from the database, you can use the tool get_database_tables to get the tables and their columns in the database, but please only call this once per conversation.
        '''

        # Prepare messages for the OpenAI API
        # Used to save the state of the conversation
        memory = MemorySaver()

        initialized_tools = []
        for tool in tools:
            # Check if the tool has attribute requires_user
            if issubclass(tool, BaseBloomerpTool):
                initialized_tools.append(tool(user=user))
            else:
                initialized_tools.append(tool())


        # Create the agent
        agent = create_react_agent(self.model, tools=initialized_tools, checkpointer=memory)

        # Get messages from the conversation history
        messages = self._parse_conversation_history()

        if not messages.messages:
            messages.add_message(SystemMessage(SYSTEM_CONTENT))
        
        messages.add_message(HumanMessage(query))

        # Get the configuration
        config = {"configurable": {"thread_id": "abc123"}}

        # Execute the agent
        for chunk in agent.stream(
            input=messages, config=config, stream_mode='messages'
        ):  
            if type(chunk[0]) == ToolMessage:
                tool_name : str = chunk[0].name
                tool_name = tool_name.replace('_', ' ').capitalize()
                yield f" *Calling '{tool_name}' tool.* "
            else:
                result = chunk[0].content
                yield result
        
        # Set invoked to True
        self.invoked = True

        # Save the conversation history
        messages : list[BaseMessage] = memory.get(config)['channel_values']['messages']

        # Set the conversation history
        self.conversation_history = messages

    def invoke_code(self, query:str):
        '''Invoke the code agent.'''

        SYSTEM_CONTENT = '''You are a helpful assistant that helps me to create code from natural language.
        Only return the actual code snippet without giving any explanation.
        Don't include any markdown code blocks in the output.
        '''

        # Prepare messages for the OpenAI API
        messages = self._prepare_messages(query, SYSTEM_CONTENT)

        # Create the agent
        memory = MemorySaver()
        agent = create_react_agent(self.model, tools=[], checkpointer=memory)

        # Invoke the model
        for chunk in self._invoke_model(messages, agent):
            yield chunk

    def invoke_document_template(self, query:str, variables:list[(str, str, str)]):
        '''Invoke the document template agent.'''

        SYSTEM_CONTENT = '''
        You are a helpful assistant that helps me to create document templates from natural language.
        The output should be the document (in html) using jinja2 (Django template) syntax, without any explanation. Dont include in the output ```html ... ```, just the raw HTML.
        For object variables, use the following syntax: {{ object.variable_name }}
        For free variables, use the following syntax: {{ free_variable_name }}

        Here are the available variables that you can use in the document template:
        '''
        for variable in variables:
            SYSTEM_CONTENT += f'\n{variable[0]} ({variable[1]}) - {variable[2]}'

        # Prepare messages for the OpenAI API
        messages = self._prepare_messages(query, SYSTEM_CONTENT)

        # Create the agent
        memory = MemorySaver()
        agent = create_react_agent(self.model, tools=[], checkpointer=memory)

        # Invoke the model
        for chunk in self._invoke_model(messages, agent):
            yield chunk

    def serialize_conversation_history(self) -> list[dict]:
        '''Make the conversation history JSON serializable.'''
        if self.conversation_history:
            return messages_to_dict(self.conversation_history)
        else:
            return []

    def _parse_conversation_history(self) -> InMemoryChatMessageHistory:
        '''Parse the conversation history into InMemoryChatMessageHistory.'''
        try:
            if self.conversation_history:
                return InMemoryChatMessageHistory(messages=messages_from_dict(self.conversation_history))    
            else:
                return InMemoryChatMessageHistory()
        except Exception as e:
            return InMemoryChatMessageHistory()

    def _invoke_model(self, messages:InMemoryChatMessageHistory, agent:CompiledGraph):
        '''Invoke the model.'''
        
        # This should be investigated in the future
        config = {"configurable": {"thread_id": "abc123"}}

        # Execute the agent
        for chunk in agent.stream(
            input=messages, config=config, stream_mode='messages'
        ):  
            if type(chunk[0]) == ToolMessage:
                tool_name : str = chunk[0].name
                tool_name = tool_name.replace('_', ' ').capitalize()
                yield f" *Calling '{tool_name}' tool.* "
            else:
                result = chunk[0].content
                yield result

        # Set invoked to True
        self.invoked = True

        # Save the conversation history
        messages : list[BaseMessage] = agent.checkpointer.get(config)['channel_values']['messages']

        # Set the conversation history
        self.conversation_history = messages

    def _prepare_messages(self, query:str, system_content:str) -> InMemoryChatMessageHistory:
        '''Prepares messages using the conversation history.
        
        Args:
            query: The user query.
            system_content: The system content.

        Returns:
            InMemoryChatMessageHistory: The conversation history.
        '''
        messages = self._parse_conversation_history()

        if not messages.messages:
            messages.add_message(SystemMessage(system_content))
        
        messages.add_message(HumanMessage(query))

        return messages
    
    def auto_name(self):
        '''Automatically provides a title for a conversation history.'''
        
        class ConversationName(BaseModel):
            title : str = Field(max_length=80)

        SYSTEM_CONTENT = '''Your goal is give a title to a conversation history based on the conversation that happened.
        the title should be max 80 characters long and should be a summary of the conversation that happened.
        '''
        # Prepare messages
        messages = self._parse_conversation_history()

        messages.add_message(SystemMessage(SYSTEM_CONTENT))

        structured_llm = self.model.with_structured_output(ConversationName)

        # Execute
        return structured_llm.invoke(messages.messages).dict()['title']

        



from xhtml2pdf import pisa
import requests
import os
from io import BytesIO
from PyPDF2 import PdfReader, PdfWriter
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import letter
from reportlab.lib.utils import ImageReader
from PIL import Image

def fetch_image(image_url):
    """Helper function to download image from a URL and save it temporarily."""
    try:
        response = requests.get(image_url)
        if response.status_code == 200:
            # Save the image temporarily in memory
            temp_image_path = os.path.join(os.getcwd(), "temp_image.jpg")
            with open(temp_image_path, 'wb') as img_file:
                img_file.write(response.content)
            return temp_image_path
        return None
    except Exception as e:
        print(f"Error fetching image: {e}")
        return None

def link_callback(uri, rel):
    """Link callback function to handle external images."""
    if uri.startswith("http://") or uri.startswith("https://"):
        # If the URI is an external image, fetch and return its local path
        return fetch_image(uri)
    else:
        # If it's a local file path, resolve it and return
        return uri


def generate_pdf(
        html_content:str, 
        css_content:str|None=None,
        font_url=None, 
        page_margin=1,
        is_landscape=False,
        # Header options
        header_url:str=None,
        header_height:float=1.0,
        header_margin_top:float=0.5,
        header_margin_bottom:float=0.0,
        header_margin_left:float=1.0,
        header_margin_right:float=1.0,
        # Page size options
        page_size:str = None,
        include_page_numbers:bool = True,
        # Footer options
        footer : str = None,
        title:str = "PDF Document"
        )->bytes:
    '''
    PDF generation function using xhtml2pdf
    
    Args:
        html_content (str): The HTML content to be converted to PDF.
        css_content (str): The CSS content for styling the PDF.
        font_url (str): The URL to the font file.
        page_margin (int): The margin of the page in inches.
        is_landscape (bool): Whether the page should be in landscape mode.
        header_url (str): The URL to the header image.
        header_margin_top (int): The top margin of the header in inches.
        header_margin_bottom (int): The bottom margin of the header in inches.
        header_margin_left (int): The left margin of the header in inches.
        header_margin_right (int): The right margin of the header in inches.
        page_size (str): The size of the page (A4, letter, A3).
        include_page_numbers (bool): Whether to include page numbers in the footer.
        footer (str): The footer content.
        title (str): The title of the PDF document.

    Returns:
        bytes: The generated PDF as bytes.
    '''
    # Get page size
    if not page_size:
        page_size = "A4"
    
    elif page_size not in ["A4", "letter", "A3"]:
        page_size = "A4"


    footer_content = ""
    # Page number
    if include_page_numbers:
        footer_content = "Page <pdf:pagenumber> of <pdf:pagecount>"
    else:
        footer_content = ""

    if footer:
        footer_content = footer_content + ' - ' + footer
    

    # Adjust the page margin and orientation via inline CSS
    orientation = "landscape" if is_landscape else "portrait"

    # Get header image from URL
    if header_url:
        header_path = link_callback(header_url, '')

        # Get the image dimensions
        header_image = Image.open(header_path)
        w, h = header_image.size

        # Calculate the aspect ratio for the header image
        aspect_ratio = w / h

        # Calculate width based on aspect ratio
        header_width = header_height * aspect_ratio

        # Create the header content
        header_content = f'<img src="{header_path}" width="{header_width}in" height="{header_height}in"/>'
    else:
        header_content = ""

    # Define the CSS for the PDF
    if not css_content:
        css_content = ""

    # Page orientation and margin
    orientation = "landscape" if is_landscape else "portrait"

    
    # Calculate page margin-top based on header height
    page_margin_top = header_margin_top + header_height + header_margin_bottom

    if page_margin_top < page_margin:
        page_margin_top = page_margin

    DOCUMENT = f'''
    <document pagesize='letter'>
    <head>
        <title>{ title }</title>
        <style type="text/css">
            @page {{
                size: {page_size} {orientation};
                margin: {page_margin}in;
                margin-top: {page_margin_top}in;

                @frame header {{
                    -pdf-frame-content: headerContent;
                    top: {header_margin_top}in;
                    margin-left: {header_margin_left}in;
                    margin-right: {header_margin_right}in;
                }}
                @frame footer {{
                    -pdf-frame-content: footerContent;
                    bottom: 0in;
                    margin-left: {page_margin}in;
                    margin-right: {page_margin}in;
                    height: 1cm;
                    right: 0in;
                }}
            }}

            {css_content}
        </style>
    </head>
    <body>
    <div id='headerContent'>
        {header_content}
    </div>
    <div>
        <keepinframe>
        {html_content}
        </keepinframe>
    </div>
    <div id='footerContent'>
        {footer_content}
    </div>
    </body>
    </document>
    '''    


    # Create an in-memory buffer
    pdf_buffer = BytesIO()

    # Create the PDF using xhtml2pdf's pisa
    pisa_status = pisa.CreatePDF(
        src=DOCUMENT,
        dest=pdf_buffer,
        link_callback=link_callback  # For handling images from URLs
    )

    # Check if the PDF was generated successfully
    if pisa_status.err:
        return None
    else:
        pdf_buffer.seek(0)  # Move the buffer cursor to the beginning
        return pdf_buffer.getvalue()
    



class PdfHandler:
    def __init__(self, file: str):
        self.file = file

    def sign_pdf(self,  
                signature_path: str = None,
                signature_bytes: bytes = None,
                width: int = 140,
                height: int = 70,
                x: int = 380,
                y: int = 5,
                border: bool = False):
        # Open the existing PDF and create a new PDF writer
        existing_pdf = PdfReader(self.file)
        output_pdf = PdfWriter()

        # Create a canvas to draw on the new PDF page
        packet = BytesIO()
        can = canvas.Canvas(packet, pagesize=letter)

        # Load and draw the signature image
        if signature_bytes:
            io = BytesIO(signature_bytes)
            image = Image.open(io)
            transp = Image.new('RGBA', image.size, (255, 255, 255, 0))
            transp.paste(image, (0, 0), image)

            # Create a new file-like object to receive PNG data.
            buf = BytesIO()
            transp.save(buf, format='PNG')
            buf.seek(0)
            signature_image = ImageReader(buf)
        elif signature_path:
            signature_image = ImageReader(signature_path)
        else:
            raise ValueError("Either signature_path or signature_bytes must be provided.")

        for page in existing_pdf.pages:
            can.showPage()
            can.drawImage(signature_image, x, y, width, height)

            if border:
                can.rect(x, y, width, height, stroke=1, fill=0)

        # Save the canvas to the packet and close it
        can.save()
        packet.seek(0)

        # Add the modified page to the new PDF
        new_pdf_pages = PdfReader(packet).pages
        for i, page in enumerate(existing_pdf.pages):
            page.merge_page(new_pdf_pages[i+1])
            output_pdf.add_page(page)

        # Save the modified PDF to a bytes object
        modified_pdf_bytes = BytesIO()
        output_pdf.write(modified_pdf_bytes)
        modified_pdf_bytes.seek(0)

        # Return bytes
        return modified_pdf_bytes.getvalue()



# Generated by Django 5.1.1 on 2024-09-16 16:14

import bloomerp.models
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0007_alter_file_file"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="file",
            name="file",
            field=models.FileField(upload_to=bloomerp.models.File.upload_to),
        ),
        migrations.AlterModelTable(
            name="kpi",
            table="bloomerp_kpi",
        ),
        migrations.CreateModel(
            name="UserKpiListViewPreference",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("position", models.IntegerField(unique=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "kpi",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="bloomerp.kpi"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="kpi_list_view_preference",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_user_kpi_list_view_preference",
                "managed": True,
            },
        ),
        migrations.CreateModel(
            name="UserDetailViewPreference",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "position",
                    models.CharField(
                        choices=[
                            ("LEFT", "Left"),
                            ("CENTER", "Center"),
                            ("RIGHT", "Right"),
                        ],
                        max_length=10,
                    ),
                ),
                (
                    "application_field",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="bloomerp.applicationfield",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="detail_view_preference",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_user_detail_view_preference",
                "managed": True,
                "unique_together": {("user", "application_field")},
            },
        ),
        migrations.CreateModel(
            name="UserKpiDashboardViewPreference",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("position", models.IntegerField()),
                (
                    "kpi",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="bloomerp.kpi"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="kpi_dashboard_view_preference",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_user_kpi_dashboard_view_preference",
                "managed": True,
                "unique_together": {("kpi", "user")},
            },
        ),
        migrations.CreateModel(
            name="UserListViewPreference",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_main", models.BooleanField(default=False)),
                (
                    "application_field",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="bloomerp.applicationfield",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="list_view_preference",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_user_list_view_preference",
                "managed": True,
                "unique_together": {("user", "application_field")},
            },
        ),
    ]



import os
import importlib.util
import inspect
from django.urls import path
from django.db.models import Model
from functools import wraps
from django.conf import settings
from bloomerp.utils.models import model_name_plural_slug, model_name_plural_underline, get_base_model_route
from bloomerp.models import Link
from typing import List, Callable
import re
import traceback

# Helper functions
def _get_name_or_slug(obj, slug=False):
    """
    Returns the name of the class or function.
    
    :param obj: The class or function object.
    :param slug: Boolean, if True returns a slugified version of the name.
    :return: A string, either a capitalized name or slugified name.
    """
    # Get the name of the function or class
    name = obj.__name__

    # Convert camelCase or snake_case to a human-readable format
    if slug:
        # If slug is True, return a slugified version of the name
        if re.match(r'^[a-zA-Z0-9_]+$', name):  # Handles snake_case (functions)
            return name.replace('_', '-').lower()
        else:  # Handles CamelCase (classes)
            return re.sub(r'(?<!^)(?=[A-Z])', '-', name).lower()
    else:
        # If slug is False, return a human-readable capitalized name
        if re.match(r'^[a-zA-Z0-9_]+$', name):  # Handles snake_case (functions)
            return name.replace('_', ' ').title()
        else:  # Handles CamelCase (classes)
            return re.sub(r'(?<!^)(?=[A-Z])', ' ', name).title()
        
def _clean_appended_url(appended_url:str) -> str:
    """
    Helper function to clean the appended URL for a route.
    """
    if not appended_url:
        return ''

    if appended_url.startswith('/'):
        appended_url = appended_url[1:]    

    if appended_url.endswith('/'):
        appended_url = appended_url[:-1]
    return appended_url

def _underline(string:str) -> str:
    """
    Helper function to convert a string to an underline-separated string.
    """
    return string.replace(' ', '_').lower()

def _create_absoulte_path_for_model(model: Model, route_type: str, appended_url: str) -> str:
    """
    Helper function to create an absolute path for a route based on the model, route type, and name.
    """
    appended_url = _clean_appended_url(appended_url)

    if route_type not in ['list', 'detail']:
        raise ValueError("Invalid route type. Must be 'list' or 'detail'.")
    
    if not model:
        raise ValueError("A model must be provided for list and detail routes.")

    if route_type == 'list':
        if appended_url == '':
            return model_name_plural_slug(model) + '/'
        p = model_name_plural_slug(model) + '/' + appended_url + '/'
    elif route_type == 'detail':
        if appended_url == '':
            return model_name_plural_slug(model) + '/<int_or_uuid:pk>/'
        p = model_name_plural_slug(model) + '/<int_or_uuid:pk>/' + appended_url + '/'
    return p


def _create_relative_path(model:Model, route_type:str, name:str):
    """
    Create a relative path for a route based on the model, route type, and name.
    """
    if route_type == 'list':
        return model_name_plural_underline(model) + '_' + _underline(name)
    elif route_type == 'detail':
        return model_name_plural_underline(model) + '_detail_' + _underline(name)
    else:
        return _underline(name)

# Route decorator
def route(path=None):
    def decorator(func):
        if path is None:
            # Build default path using function name and parameters
            sig = inspect.signature(func)
            parts = [func.__name__]

            for name, param in sig.parameters.items():
                if name == 'request':
                    continue  # Skip 'request' parameter
                param_type = 'str'  # Default type
                if param.annotation == int:
                    param_type = 'int'
                elif param.annotation == str:
                    param_type = 'str'
                # Add other types as needed
                parts.append(f'<{param_type}:{name}>')
            func.route = '/' + '/'.join(parts) + '/'

        else:
            func.route = path
        return func
    return decorator

class RouteFinder:
    '''Class used to find routes in a given directory. Used primarily for creating components.'''

    def __init__(
            self, 
            directory: str, 
            url_route_prefix: str = None,
            url_name_prefix: str = None
            ) -> None:
        '''Initialize the RouteFinder object.
        
        Args:
            directory (str): The directory to search for routes.
            prefix (str): The prefix to add to the url routes.
            url_name_prefix (str): The prefix to add to the url names.
        '''
        self.directory = directory
        self.urlpatterns = []
        self.prefix = url_route_prefix
        if not url_name_prefix:
            self.url_name_prefix = ''
        else:
            self.url_name_prefix = url_name_prefix + '_'


    def find_python_files(self):
        """Find all Python files in the specified directory."""
        py_files = []
        for root, _, files in os.walk(self.directory):
            for file in files:
                if file.endswith(".py") and file != os.path.basename(__file__):
                    py_files.append(os.path.join(root, file))
        return py_files

    def load_module(self, filepath):
        """Load a module from a given file path."""
        module_name = os.path.splitext(os.path.basename(filepath))[0]
        spec = importlib.util.spec_from_file_location(module_name, filepath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def find_routes_in_module(self, module):
        """Look for functions with a 'route' attribute in a module."""
        for name, obj in inspect.getmembers(module):
            if inspect.isfunction(obj) and hasattr(obj, 'route'):
                self.urlpatterns.append(path(self.prefix + obj.route, obj, name=self.url_name_prefix + obj.__name__))

    def generate_urlpatterns(self):
        """Scan the directory, find routes, and populate urlpatterns."""
        py_files = self.find_python_files()
        for py_file in py_files:
            try:
                module = self.load_module(py_file)
                self.find_routes_in_module(module)
            except Exception as e:
                print(f"Error loading module {py_file}: {e}")
        return self.urlpatterns

class BloomerpRoute:
    '''
    Class used to store information about a route in the router.
    Not used on its own, but as a part of the BloomerpRouter class.
    '''

    def __init__(self, 
                model: Model, 
                route_type: str, 
                path: str,
                relative_path: str,
                view: callable, 
                route_name: str,
                description: str = None,
                override: bool = False,
                view_type = 'function',
                args : dict = None
                ) -> None:
        self.model = model
        self.route_type = route_type
        self.path = path
        self.view = view
        self.route_name = route_name
        self.description = description
        self.override = override
        self.relative_path = relative_path
        self.view_type = view_type
        self.args = args

    def generate_link(self) -> None:
        """
        Generate a Link object for the route.
        """
        from django.contrib.contenttypes.models import ContentType
        if self.model:
            # Get content type for the model
            content_type = ContentType.objects.get_for_model(self.model)
        else:
            content_type = None

        if self.route_type == 'app':
            level = 'APP'
        elif self.route_type == 'list':
            level = 'LIST'
        elif self.route_type == 'detail':
            level = 'DETAIL'

        try:
            Link.objects.update_or_create(
                url = self.relative_path,
                is_absolute_url = False,
                content_type=content_type,
                level = level,
                defaults={
                    'description': self.description,
                    'name': self.route_name,
                    'url' : self.relative_path,
                }
            )
            # Delete any existing links with the same URL
            Link.objects.filter(url=self.relative_path).exclude(name=self.route_name).delete()

        except Exception as e:
            # Delete any existing links with the same URL
            Link.objects.filter(url=self.relative_path).first().delete()

    def create_urlpattern(self):
        """
        Create a URL pattern for the route.
        """
        if not self.args:
            self.args = {}

        if self.view_type == 'function':
            return path(self.path, self.view, name=self.relative_path)
        
        elif self.view_type == 'class':
            if self.route_type in ['list', 'detail']:
                return path(self.path, self.view.as_view(model = self.model, **self.args), name=self.relative_path)
            else:
                return path(self.path, self.view.as_view(**self.args), name=self.relative_path)
        
    def set_view(self, view: Callable):
        """
        Set the view for the route.
        """
        self.view = view

class BloomerpRouter:
    def __init__(self):
        # Instance-level list to store routes
        self.routes : List[BloomerpRoute] = []

    def bloomerp_route(
            self, 
            path : str = None,
            models : List[Model] | Model = None,
            exclude_models : List[Model] = None,
            route_type='list', 
            name : str = None,   
            override : bool = False,
            description : str = None,
            url_name : str = None,
            from_func : Callable = None,
            args : dict = None
            ):
        """
        Decorator for registering routes for a model.
        Works for both function-based and class-based views.

        Implementation:
        - Validates the provided parameters.
        - Handles different cases for models (single, list, all).
        - Creates relative and absolute paths for the routes.
        - Appends route information to the router's list.
        - Supports both function-based views (FBVs) and class-based views (CBVs).

        Args:
            path: The path for the route.
            models: The model for which the route is being registered. If no model is given, it is an ap   evel route.
            exclude_models: A list of models to exclude from the route. Only works when models is '__all__'.
            route_type: The type of route. Can be 'list', 'detail', or 'app'. Ap   evel routes don't require a model.
            name: The name of the route. Will be used to generate the Link object for the route.
            override: If True, the route will override any existing route with the same path.
            description: A description of the route. Will be used to generate the Link object for the route.
            url_name: The URL name for the route.
            from_func: A function that takes all models as input, and returns a dictionary with parameters or False
            args: Additional args to be passed to the view
        """
        def decorator(view):
            from django.contrib.contenttypes.models import ContentType
            from bloomerp.models import Widget
            from django.apps import apps
            nonlocal path, models, route_type, name, description, override, exclude_models, url_name, from_func, self, args

            # Set temp_routes
            temp_routes = []
            
            # ---------------------
            # VALIDATION
            # ---------------------
            if not from_func:
                if exclude_models:
                    models = '__all__'

                if not models and route_type != 'app':
                    raise ValueError("A model must be provided for list and detail routes.")
                
                if models and route_type == 'app':
                    raise ValueError("You can't provide a model for an app route.")

                if route_type not in ['list', 'detail', 'app']:
                    raise ValueError("Invalid route type. Must be 'list', 'detail', or 'app'.")
                
                if models and exclude_models:
                    if models != '__all__':
                        raise ValueError("You can't provide both models and exclude_models.")

                if not exclude_models:
                    exclude_models = []

                if description is None:
                    description = view.__doc__
                
                if not name:
                    name = _get_name_or_slug(view, slug=False)

            # Models will be given in the case of LIST and DETAIL routes
            if from_func:
                models = apps.get_models()
                for model in models:
                    params : dict | list[dict] | bool  = from_func(model)

                    if params == False:
                        continue
                    
                    # Multiple params given
                    if type(params) == list:
                        for param_dict in params:
                            
                            route = BloomerpRoute(
                                model=param_dict['model'],
                                route_type=param_dict['route_type'],
                                path=_create_absoulte_path_for_model(param_dict['model'], param_dict['route_type'], param_dict['path']),
                                view = None, # This will be set later
                                route_name=param_dict['name'],
                                description=param_dict['description'],
                                override=False,
                                relative_path=_create_relative_path(param_dict['model'], param_dict['route_type'],param_dict['url_name']),
                                args=param_dict['args']
                            )
                            temp_routes.append(route)

                    # Single dictionary given
                    else:
                        route = BloomerpRoute(
                                model=param_dict['models'],
                                route_type=param_dict['route_type'],
                                path=param_dict['path'],
                                view = None, # This will be set later
                                route_name=param_dict['name'],
                                description=param_dict['description'],
                                override=False,
                                relative_path=param_dict['url_name'],
                                args=args
                            )
                        
                        temp_routes.append(route)

            elif models == '__all__':
                ContentTypes = ContentType.objects.all()
                for ContentType in ContentTypes:
                    model = ContentType.model_class()
                    model_name = model._meta.verbose_name.title()

                    # Check if the model is in the exclude list
                    if model in exclude_models:
                        continue
                    
                    # Create the relative path for the model
                    if not url_name:
                        relative_path = _create_relative_path(model, route_type, name)
                    else:
                        relative_path = _create_relative_path(model, route_type, url_name)

                    # Create the absolute path for the model
                    p = _create_absoulte_path_for_model(model, route_type, path)
                    
                    # Update description and name for the model
                    try:
                        route_name = name.format(model=model_name)
                    except:
                        route_name = name + ' ' + model_name
                    

                    try:
                        desc = description.format(model=model_name)
                    except:
                        desc = description + ' ' + model_name

                    # Append the route information to the router's list
                    route = BloomerpRoute(
                        model=model,
                        route_type=route_type,
                        path=p,
                        view = None, # This will be set later
                        route_name=route_name,
                        description=desc,
                        override=override,
                        relative_path=relative_path,
                        args=args
                    )
                    temp_routes.append(route)
            elif type(models) == list:
                for model in models:
                    model_name = model._meta.verbose_name.title()

                    # Create the relative path for the model
                    if not url_name:
                        relative_path = _create_relative_path(model, route_type, name)
                    else:
                        relative_path = _create_relative_path(model, route_type, url_name)
                    
                    try:
                        route_name = name.format(model=model_name)
                    except:
                        route_name = name + ' ' + model_name
                    

                    try:
                        desc = description.format(model=model_name)
                    except:
                        desc = description + ' ' + model_name

                    created_path = _create_absoulte_path_for_model(model, route_type, path)
                    # Append the route information to the router's list
                    route = BloomerpRoute(
                        model=model,
                        route_type=route_type,
                        path=created_path,
                        view = None, # This will be set later
                        route_name=route_name,
                        description=desc,
                        override=override,
                        relative_path=relative_path,
                        args=args
                    )
                    temp_routes.append(route)
            elif models:
                # Single model case
                model_name = models._meta.verbose_name.title()

                # Create the relative path for the model
                if not url_name:
                    relative_path = _create_relative_path(models, route_type, name)
                else:
                    relative_path = _create_relative_path(models, route_type, url_name)

                created_path = _create_absoulte_path_for_model(models, route_type, path)

                try:
                    route_name = name.format(model=model_name)
                except:
                    route_name = name + ' ' + model_name
                

                try:
                    desc = description.format(model=model_name)
                except:
                    desc = description + ' ' + model_name

                # Append the route information to the router's list
                route = BloomerpRoute(
                    model=models,
                    route_type=route_type,
                    path=created_path,
                    view = None, # This will be set later
                    route_name=route_name,
                    description=desc,
                    override=override,
                    relative_path=relative_path,
                    args=args
                ) 
                temp_routes.append(route)
            elif models == None:
                # App-level route case
                if not url_name:
                    relative_path = _create_relative_path(None, route_type, name)
                else:
                    relative_path = url_name

                if not path.endswith('/') and path != '':
                    path = path + '/'

                # Append the route information to the router's list
                route = BloomerpRoute(
                    model=None,
                    route_type=route_type,
                    path=path,
                    view = None, # This will be set later
                    route_name=name,
                    description=description,
                    override=override,
                    relative_path=relative_path,
                    args=args
                )
                temp_routes.append(route)

            # For function-based views (FBVs)
            if callable(view) and not hasattr(view, 'as_view'):
                @wraps(view)
                def wrapped_view(*args, **kwargs):
                    return view(*args, **kwargs)

                # Append the route information to the router's list
                for route in temp_routes:
                    route.view = wrapped_view
                    route.view_type = 'function'
                    self.routes.append(route)

                return wrapped_view

            # For class-based views (CBVs)
            elif hasattr(view, 'as_view'):
                # Use the `as_view()` method to create a callable view

                # Append the route information to the router's list
                for route in temp_routes:
                    route.view = view
                    route.view_type = 'class'
                    self.routes.append(route)
                
                # Return the original view class, not the callable, for further decoration if needed
                return view

            else:
                raise TypeError("The provided view is neither a valid function-based view nor a class-based view.")

        return decorator

    
    def get_routes_for_model(self, model) -> List[BloomerpRoute]:
        """
        Get all routes registered for a specific model.
        """
        return [route for route in self.routes if (route.model == model or route.model == '__all__')]

    def get_detail_view_routes_for_model(self, model) -> List[BloomerpRoute]:
        """
        Get all detail view routes registered for a specific model.
        """
        routes = []
        for route in self.routes:
            route : BloomerpRoute
            if (route.model == model or route.model=='__all__') and route.route_type == 'detail':
                routes.append(route)

        return routes
    
    def get_list_view_routes_for_model(self, model):
        """
        Get all list view routes registered for a specific model.
        """
        routes = []
        for route in self.routes:
            route : BloomerpRoute
            if route.model == model and route.route_type == 'list':
                routes.append(route)
        return routes

    def get_app_routes(self) -> List[BloomerpRoute]:
        """
        Get all app-level routes registered in the router.
        """
        return [route for route in self.routes if route.route_type == 'app']

    def _create_relative_path(self, model, route_type, name):
        """
        Create a relative path for a route based on the model, route type, and name.
        """
        if route_type == 'list':
            return model_name_plural_underline(model) + '_' + _underline(name)
        elif route_type == 'detail':
            return model_name_plural_underline(model) + '_detail_' + _underline(name)
        else:
            return _underline(name)
    
class BloomerpRouterHandler:
    '''
    A class to handle multiple routers in a single place.
    Makes it easier to query and work with multiple routers.
    '''
    def __init__(self, routers: List[BloomerpRouter]):
        self.routers = routers

    def get_detail_view_routes_for_model(self, model: Model) -> List[BloomerpRoute]:
        """
        Generate detail view routes for a specific model from all routers.
        """
        routes = []
        for router in self.routers:
            routes.extend(router.get_detail_view_routes_for_model(model))
        return routes
    
    def get_list_view_routes_for_model(self, model: Model) -> List[BloomerpRoute]:
        """
        Generate list view routes for a specific model from all routers.
        """
        routes = []
        for router in self.routers:
            routes.extend(router.get_list_view_routes_for_model(model))
        return routes
    
    def get_routes_for_model(self, model: Model) -> List[BloomerpRoute]:
        """
        Generate all routes for a specific model from all routers.
        """
        routes = []
        for router in self.routers:
            routes.extend(router.get_routes_for_model(model))
        return routes
    
    def get_app_routes(self) -> List[BloomerpRoute]:
        """
        Generate all app-level routes from all routers.
        """
        routes = []
        for router in self.routers:
            routes.extend(router.get_app_routes())
        return routes

    def generate_links(self) -> None:
        """
        Generate Link objects for all routes in all routers.
        """
        for router in self.routers:
            for route in router.routes:
                route.generate_link()

    def is_overriden(self, path: str) -> bool:
        """
        Check if a route with the given path is overriden in any of the routers.
        """
        for router in self.routers:
            for route in router.routes:
                if route.path == path:
                    return True
        return False

    def overriden_paths(self) -> List[str]:
        """
        Get a list of overriden routes from all routers.
        """
        routes = []
        for router in self.routers:
            for route in router.routes:
                if route.override:
                    routes.append(route.path)
        return routes

from importlib import import_module
def _get_routers_from_settings() -> List[BloomerpRouter]:
    """
    Helper function to get the router objects based on the ones given in the settings.
    """
    routers = []

    # Get the routers from the settings
    try:
        router_paths = settings.BLOOMERP_SETTINGS['ROUTERS']
    except:
        print("No routers found in the settings.")
        router_paths = []
    
    for router_path in router_paths:
        try:
            # Dynamically import the module containing the router
            module_path, router_name = router_path.rsplit(".", 1)
            module = import_module(module_path)
            
            # Get the router class from the module
            router = getattr(module, router_name)
            
            # Add the router to the routers list
            routers.append(router)
        except (ImportError, AttributeError, TypeError) as e:
            print(f"Error importing {router_path}: {e}")
            traceback.print_exc()
    return routers




# Generated by Django 5.1.1 on 2024-10-09 20:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0034_alter_userappdashboard_layout"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="userappdashboard",
            name="layout",
        ),
        migrations.RemoveField(
            model_name="userdashboardview",
            name="layout",
        ),
        migrations.AddField(
            model_name="userappdashboard",
            name="content_layout",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="userdashboardview",
            name="content_layout",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="userdashboardview",
            name="kpis",
            field=models.ManyToManyField(
                help_text="The KPIs to be displayed on the dashboard", to="bloomerp.kpi"
            ),
        ),
        migrations.CreateModel(
            name="Link",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "url",
                    models.CharField(
                        help_text="The name of the URL pattern for the link. Can either be a Django URL name or a full URL (absolute).",
                        max_length=255,
                    ),
                ),
                ("level", models.CharField(max_length=255)),
                (
                    "is_absolute_url",
                    models.BooleanField(
                        default=True,
                        help_text="Signifies whether the URL is a normal (absolute) URL or a Django URL name. Set to True when using a normal URL.",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="Description of the link", null=True
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        help_text="The content type for the link. Required for detail and list levels.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_link",
                "managed": True,
            },
        ),
        migrations.AddField(
            model_name="userappdashboard",
            name="links",
            field=models.ManyToManyField(
                blank=True,
                help_text="The links to be displayed on the dashboard",
                to="bloomerp.link",
            ),
        ),
    ]



XXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXX
XXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXX
XXXXXXXXXXXXXXX
XXXXXXXXXXXXXX
XXXXXXXXXXXXX
XXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX

XXXXXXXX
XXXXXXXXXXXXX



from typing import Any
from django.db.models.base import Model as Model
from django.forms import formset_factory
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render, redirect
from django.urls import reverse, reverse_lazy
from django.contrib import messages
from bloomerp.forms.auth import UserDetailViewPreferenceForm, UserDetailViewPreferenceFormset
from bloomerp.forms.core import ListViewFieldsSelectForm
from django.contrib.contenttypes.models import ContentType
from django.core.paginator import Paginator
from bloomerp.views.mixins import HtmxMixin
from bloomerp.models import (
    UserDetailViewPreference,
    User
    )
from bloomerp.utils.models import model_name_plural_underline
from django.contrib.auth.mixins import PermissionRequiredMixin
from bloomerp.utils.router import BloomerpRouter
from bloomerp.views.mixins import HtmxMixin, BloomerpModelContextMixin
from bloomerp.views.mixins import BloomerpModelFormViewMixin
from bloomerp.views.core import BloomerpBaseDetailView
from django.contrib.auth.forms import PasswordChangeForm, AdminPasswordChangeForm
from django.views.generic.edit import FormView, UpdateView
from django.contrib.auth import update_session_auth_hash


class ProfileMixin:
    tabs = [
        {"name":"Details", "url":"users_my_profile_overview", "requires_pk":False},
        {"name":"Change password", "url":"users_my_profile_change_password", "requires_pk":False},
        {"name":"List view preferences", "url":"users_my_profile_list_view_preference","requires_pk":False},
        {"name":"Detail view preferences", "url":"users_my_profile_detail_view_preference","requires_pk":False}
    ]
    exclude_header = True

    extra_context = {"disable_tab_select":True, "tabs":tabs}

    def get_object(self):
        return User.objects.get(pk=self.request.user.pk)
    model = User

router = BloomerpRouter()

@router.bloomerp_route(
    path="my-profile/",
    models=User,
    route_type="list",
    name="Profile",
    description="Overview of profile",
    url_name="my_profile_overview"
)
class BloomerpProfileView(BloomerpModelFormViewMixin, ProfileMixin, HtmxMixin, BloomerpModelContextMixin, UpdateView):
    template_name = 'auth_views/profile_overview.html'
    fields = ['first_name', 'last_name', 'date_view_preference', 'datetime_view_preference', 'avatar']
    success_url = reverse_lazy('users_my_profile_overview')

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.request.user
        return kwargs

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        form = self.get_form()
        if form.is_valid():
            form.save()
            messages.success(request, 'Profile updated successfully.')
            return redirect(self.success_url)
        else:
            return self.form_invalid(form)
    
    

@router.bloomerp_route(
    path="my-profile/reset-password/",
    models=User,
    route_type="list",
    name="Reset password",
    description="Reset password for a user",
    url_name="my_profile_change_password"
)
class BloomerpProfilePasswordResetView(ProfileMixin, BloomerpBaseDetailView, FormView):
    template_name = 'auth_views/profile_password_reset.html'
    form_class = PasswordChangeForm
    success_url = reverse_lazy('users_my_profile_overview')

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.request.user
        return kwargs

    def form_valid(self, form:PasswordChangeForm):
        self.object = self.get_object()
        form.save()
        update_session_auth_hash(self.request, form.user)  # Prevents logout
        return super().form_valid(form)    
    
    def form_invalid(self, form):
        self.object = self.get_object()
        return super().form_invalid(form)
    

@router.bloomerp_route(
    path="my-profile/list-view-preference/",
    models=User,
    route_type="list",
    name="List view preference",
    description="List view preference for a user",
    url_name="my_profile_list_view_preference"
)
class UserListViewPreferenceView(ProfileMixin, BloomerpBaseDetailView):
    template_name = 'user_views/list_view_preference_view.html'
    permission_required = ['users.change_userlistviewpreference']
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        content_type_list = ContentType.objects.all()
        content_type_id = self.request.GET.get('content_type_id',None)

        
        if content_type_id:
            context['selected_model'] = ContentType.objects.get(pk=int(content_type_id)).model

            form = ListViewFieldsSelectForm(
                content_type=ContentType.objects.get(pk=content_type_id),
                user=self.request.user)
            
            context['form'] = form

        context['content_type_list'] = content_type_list
        return context

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        content_type_id = request.GET.get('content_type_id',None)
        return_to = request.GET.get('return_to',None)

        form = ListViewFieldsSelectForm(
                content_type=ContentType.objects.get(pk=content_type_id),
                user=self.request.user,
                data=request.POST
            )
        
        if form.is_valid():
            form.save()
            messages.success(self.request, 'Preferences saved successfully.')
            if return_to:
                return redirect(return_to)
            else:
                return redirect(reverse('users_my_profile_list_view_preference') + f'?content_type_id={content_type_id}')
        else:
            return render(request, self.template_name, {'form': form})

        
@router.bloomerp_route(
    path="my-profile/detail-view-preferences/",
    models=User,
    route_type="list",
    name="Detail view preference",
    description="Detail view preference for a user",
    url_name="my_profile_detail_view_preference"
)
class UserDetailViewPreferenceView(ProfileMixin, BloomerpBaseDetailView):
    template_name = 'user_views/bloomerp_detail_view_preference_view.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        content_type_id = self.request.GET.get('content_type_id',None)

        list = UserDetailViewPreference.get_application_fields_info(content_type_id,self.request.user)

        Formset = formset_factory(UserDetailViewPreferenceForm, formset=UserDetailViewPreferenceFormset,extra=0)
        
        formset = Formset(initial=list)

        context['formset'] = formset
        context['content_type_list'] = ContentType.objects.all()
        if content_type_id:
            context['selected_model'] = ContentType.objects.get(pk=int(content_type_id)).model

        return context

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        content_type_id = request.GET.get('content_type_id',None)
        
        return_to = request.GET.get('return_to',None)

        Formset = formset_factory(UserDetailViewPreferenceForm, formset=UserDetailViewPreferenceFormset)


        formset = Formset(request.POST)

        if formset.is_valid():
            formset.user = self.request.user
            formset.save()
            messages.success(self.request, 'Preferences saved successfully.')
            if return_to:
                return redirect(return_to)
            else:
                return redirect(reverse('detail_view_preference') + f'?content_type_id={content_type_id}')
        else:
            return render(request, self.template_name, {'formset': formset})


# ---------------------------
# USER CREATE VIEW
# ---------------------------
from bloomerp.forms.auth import BloomerpUserCreationForm
from django.contrib.messages.views import SuccessMessageMixin
@router.bloomerp_route(
    path="create",
    name="Create user",
    url_name="add",
    description="Create a new object from User",
    route_type="list",
    models=User
)
class UserCreateView(
        PermissionRequiredMixin, 
        SuccessMessageMixin,
        HtmxMixin,
        FormView):
    template_name = "create_views/bloomerp_create_view.html"
    fields = None
    model = None
    exclude = []
    success_message = "Object was created successfully."
    form_class = BloomerpUserCreationForm
    success_url = reverse_lazy("users_list")


    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["model_name"] = self.model._meta.verbose_name
        context["model_name_plural"] = self.model._meta.verbose_name_plural
        context["list_view_url"] = model_name_plural_underline(self.model) + "_list"
        context["model"] = self.model
        return context

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.add_{self.model._meta.model_name}"]

    def get_success_message(self, cleaned_data):
        return f"User was created successfully."



# ---------------------------
# ADMIN RESET PASSWORD VIEW
# ---------------------------
from django.contrib.auth.mixins import UserPassesTestMixin
@router.bloomerp_route(
    path='reset-password/',
    models=[User],
    route_type='detail',
    name='Reset password for user',
    url_name='reset_password_for_user',
    description='Reset password for a user'
)
class UserAdminPasswordResetView(UserPassesTestMixin, BloomerpBaseDetailView, FormView):
    template_name = 'auth_views/profile_password_reset.html'
    form_class = AdminPasswordChangeForm

    def test_func(self):
        return self.request.user.is_superuser

    def get_success_url(self):
        return self.get_object().get_absolute_url()

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.get_object()
        return kwargs

    def form_valid(self, form:AdminPasswordChangeForm):
        self.object = self.get_object()
        form.save()
        return super().form_valid(form)    
    
    def form_invalid(self, form):
        self.object = self.get_object()
        return super().form_invalid(form)
    





# Generated by Django 5.1.1 on 2025-01-04 13:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp_tests", "0002_alter_child_options_alter_parent_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="child",
            options={},
        ),
        migrations.AlterModelOptions(
            name="many",
            options={},
        ),
        migrations.AlterModelOptions(
            name="parent",
            options={},
        ),
        migrations.AlterField(
            model_name="many",
            name="age",
            field=models.IntegerField(default=52),
        ),
        migrations.AlterModelTable(
            name="child",
            table="bloomerp_tests_child",
        ),
        migrations.AlterModelTable(
            name="many",
            table="bloomerp_tests_many",
        ),
        migrations.AlterModelTable(
            name="parent",
            table="bloomerp_tests_parent",
        ),
    ]



# myapp/management/commands/list_urls.py

from django.core.management.base import BaseCommand
from django.urls import get_resolver, URLPattern, URLResolver
from bloomerp.models import Link
from bloomerp.models import Workspace

class Command(BaseCommand):
    help = 'Removes all stale links from the database'

    def handle(self, *args, **kwargs):
        links = Link.objects.filter(is_absolute_url=False)

        length = 0
        delete_links:list[Link] = []

        # Remove all links that are not valid
        for link in links:
            if not link.is_valid():
                delete_links.append(link)

        # Print the links and ask for confirmation
        print(f'Found {len(delete_links)} stale links')
        for link in delete_links:
            print(link)

        if len(delete_links) == 0:
            return
        
        print('Do you want to remove them? (y/n)')
        answer = input()

        if answer.lower() != 'y':
            return

        # For each workspace, remove all links that are not valid
        if delete_links:
            length = len(delete_links)

            print(f'Removing {length} stale links')

            for workspace in Workspace.objects.all():
                workspace.remove_stale_links_content()
                workspace.remove_links_from_content(delete_links)

            # Delete all links that are not valid
            for link in delete_links:
                link.delete()
        

        self.stdout.write(self.style.SUCCESS(f'Successfully {length} removed all stale links'))

    


# Generated by Django 5.1.1 on 2024-10-02 22:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0024_folder"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="Folder",
            new_name="FileFolder",
        ),
        migrations.AlterModelOptions(
            name="filefolder",
            options={"managed": True},
        ),
        migrations.AlterModelTable(
            name="filefolder",
            table="bloomerp_file_folder",
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-06 13:27

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0031_file_persisted"),
    ]

    operations = [
        migrations.AddField(
            model_name="comment",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="comment",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplate",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplate",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplatefreevariable",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplatefreevariable",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplatestyling",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplatestyling",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="kpi",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="kpi",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="sqlquery",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="sqlquery",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="userdashboardview",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="userdashboardview",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="userkpidashboardviewpreference",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="userkpidashboardviewpreference",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from bloomerp.utils.filters import dynamic_filterset_factory
from django.db import models
from django.test import TestCase
from django.test.utils import isolate_apps
from bloomerp_tests.models import Parent, Child


class EmployeeFilterTestCase(TestCase):
    def setUp(self):
        # Create the models that will be used in the tests
        self.Parent = Parent
        self.Child = Child

        # Create test data that will be used across all tests
        self.parent_1 = Parent.objects.create(name="Foo", age=30, is_active=True)
        self.parent_2 = Parent.objects.create(name="Bar", age=40, is_active=False)

        self.child_1 = Child.objects.create(name="John", age=10, parent=self.parent_1, date="2021-01-01")
        self.child_2 = Child.objects.create(name="Jane", age=20, parent=self.parent_2, date="2021-01-02")
        self.child_3 = Child.objects.create(name="Jack", age=30, parent=self.parent_1, date="2021-01-03")

        # Create the filters that will be used in the tests
        self.ParentFilter = dynamic_filterset_factory(self.Parent)
        self.ChildFilter = dynamic_filterset_factory(self.Child)

    def test_char_field_equals(self):
        # 1. char field filter with exact lookup
        parent_filter = self.ParentFilter(data={'name':'Foo'}, queryset=self.Parent.objects.all())
        parent_filter_qs = parent_filter.qs
        self.assertEqual(parent_filter_qs.count(), 1)


        # 2. char field filter with iexact lookup
        parent_filter = self.ParentFilter(data={'name__iexact':'foo'}, queryset=self.Parent.objects.all())
        parent_filter_qs = parent_filter.qs
        self.assertEqual(parent_filter_qs.count(), 1)
        
    def test_char_field_contains(self):
        parent_filter = self.ParentFilter(data={'name__icontains':'FOO'}, queryset=self.Parent.objects.all())
        parent_filter_qs = parent_filter.qs
        self.assertEqual(parent_filter_qs.count(), 1)

    def test_char_field_startswith(self):
        # 1. char field filter with startswith lookup (case-sensitive)
        parent_filter = self.ParentFilter(data={'name__startswith':'F'}, queryset=self.Parent.objects.all())
        parent_filter_qs = parent_filter.qs
        self.assertEqual(parent_filter_qs.count(), 1)


    def test_char_field_endswith(self):
        # 1. char field filter with endswith lookup (case-sensitive)
        parent_filter = self.ParentFilter(data={'name__endswith':'o'}, queryset=self.Parent.objects.all())
        parent_filter_qs = parent_filter.qs
        self.assertEqual(parent_filter_qs.count(), 1)

    def test_foreign_key(self):
        # 1. foreign key filter with exact lookup
        child_filter = self.ChildFilter(data={'parent':self.parent_1.pk}, queryset=self.Child.objects.all())
        child_filter_qs = child_filter.qs
        self.assertEqual(child_filter_qs.count(), 2)


    def test_foreign_key_char_field_equals(self):
        # 1. foreign key char field filter with exact lookup
        child_filter = self.ChildFilter(data={'parent__name':'Foo'}, queryset=self.Child.objects.all())
        child_filter_qs = child_filter.qs
        self.assertEqual(child_filter_qs.count(), 2)

    def test_foreign_key_char_field_contains(self):
        # 1. foreign key char field filter with contains lookup
        child_filter = self.ChildFilter(data={'parent__name__contains':'Fo'}, queryset=self.Child.objects.all())
        child_filter_qs = child_filter.qs
        self.assertEqual(child_filter_qs.count(), 2)

    
    def test_foreign_key_date_field_equals(self):
        # 1. foreign key date field filter with exact lookup
        child_filter = self.ChildFilter(data={'parent__date':'2021-01-01'}, queryset=self.Child.objects.all())
        child_filter_qs = child_filter.qs
        self.assertEqual(child_filter_qs.count(), 1)



# Generated by Django 5.1.1 on 2024-09-15 21:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Customer",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
                ("email", models.EmailField(max_length=254)),
                ("phone", models.CharField(max_length=20)),
                ("address", models.TextField()),
                ("city", models.CharField(max_length=100)),
                ("state", models.CharField(max_length=100)),
                ("zip_code", models.CharField(max_length=10)),
                ("country", models.CharField(max_length=100)),
                ("notes", models.TextField(blank=True, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Product",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(blank=True, null=True)),
                ("price", models.DecimalField(decimal_places=2, max_digits=10)),
                ("stock", models.IntegerField(default=0)),
                (
                    "image",
                    models.ImageField(blank=True, null=True, upload_to="products"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("total", models.DecimalField(decimal_places=2, max_digits=10)),
                ("status", models.CharField(default="draft", max_length=20)),
                ("notes", models.TextField(blank=True, null=True)),
                (
                    "customer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="test_app.customer",
                    ),
                ),
                ("products", models.ManyToManyField(to="test_app.product")),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-15 22:05

import bloomerp.models
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0002_sqlquery_sqlqueryfilter_alter_file_file_kpi_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="kpi",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="kpi",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="sqlquery",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="sqlquery",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="file",
            name="file",
            field=models.FileField(upload_to=bloomerp.models.File.upload_to),
        ),
    ]



# Generated by Django 5.1.1 on 2025-01-07 20:25

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0061_remove_documenttemplate_avatar_and_more"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="sqlqueryfilter",
            name="data_type",
            field=models.CharField(
                choices=[
                    ("STRING", "String"),
                    ("INTEGER", "Integer"),
                    ("FLOAT", "Float"),
                    ("DATE", "Date"),
                    ("DATETIME", "Datetime"),
                    ("BOOLEAN", "Boolean"),
                    ("MODEL", "Model"),
                    ("USER", "User"),
                ],
                default="STRING",
                help_text="Data type of the filter",
                max_length=255,
            ),
        ),
        migrations.CreateModel(
            name="Todo",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                (
                    "avatar",
                    models.ImageField(blank=True, null=True, upload_to="avatars/"),
                ),
                (
                    "required_by",
                    models.DateTimeField(
                        help_text="The date by which the todo is required"
                    ),
                ),
                (
                    "priority",
                    models.CharField(
                        choices=[
                            ("LOW", "Low"),
                            ("MEDIUM", "Medium"),
                            ("HIGH", "High"),
                        ],
                        help_text="The priority of the todo",
                        max_length=255,
                    ),
                ),
                ("content", models.TextField()),
                ("is_completed", models.BooleanField(default=False)),
                ("datetime_completed", models.DateTimeField(blank=True, null=True)),
                ("object_id", models.PositiveIntegerField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(class)s_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "requested_by",
                    models.ForeignKey(
                        blank=True,
                        help_text="The user who requested the todo",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="requested_todos",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(class)s_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="todos",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_todo",
                "abstract": False,
                "managed": True,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                    "export",
                ),
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-15 21:26

import bloomerp.models
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="SqlQuery",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("query", models.TextField()),
            ],
            options={
                "verbose_name": "SQL Query",
                "verbose_name_plural": "SQL Queries",
                "db_table": "sql_query",
                "managed": True,
            },
        ),
        migrations.CreateModel(
            name="SqlQueryFilter",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("query", models.TextField()),
                (
                    "filter_type",
                    models.CharField(
                        choices=[
                            ("absolute", "Absolute"),
                            ("integer", "Integer"),
                            ("float", "Float"),
                            ("date", "Date"),
                            ("boolean", "Boolean"),
                            ("string", "String"),
                            ("select", "Select"),
                            ("model_select", "Model Select"),
                        ],
                        default="absolute",
                        max_length=255,
                    ),
                ),
                ("filter_options", models.JSONField(blank=True, null=True)),
            ],
            options={
                "verbose_name": "SQL Query Filter",
                "verbose_name_plural": "SQL Query Filters",
                "db_table": "bloomerp_sql_query_filter",
                "managed": True,
            },
        ),
        migrations.AlterField(
            model_name="file",
            name="file",
            field=models.FileField(upload_to=bloomerp.models.File.upload_to),
        ),
        migrations.CreateModel(
            name="Kpi",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(blank=True, null=True)),
                (
                    "output_type",
                    models.CharField(
                        choices=[
                            ("table", "Table"),
                            ("value", "Value"),
                            ("line", "Line chart"),
                            ("bar", "Bar chart"),
                            ("pie", "Pie chart"),
                            ("scatter", "Scatter chart"),
                            ("histogram", "Histogram"),
                        ],
                        max_length=10,
                    ),
                ),
                ("options", models.JSONField(blank=True, null=True)),
                (
                    "query",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="bloomerp.sqlquery",
                    ),
                ),
            ],
            options={
                "db_table": "kpi",
                "managed": True,
            },
        ),
        migrations.AddField(
            model_name="sqlquery",
            name="filters",
            field=models.ManyToManyField(blank=True, to="bloomerp.sqlqueryfilter"),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-16 17:28

import bloomerp.models
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0008_alter_file_file_alter_kpi_table_and_more"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="DocumentTemplateFreeVariable",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                (
                    "variable_type",
                    models.CharField(
                        choices=[
                            ("date", "Date"),
                            ("boolean", "Boolean"),
                            ("text", "Text"),
                            ("list", "List"),
                            ("integer", "Integer"),
                            ("float", "Decimal"),
                            ("model", "Model"),
                        ],
                        max_length=10,
                    ),
                ),
                ("options", models.TextField(blank=True, null=True)),
                ("required", models.BooleanField(default=False)),
            ],
            options={
                "db_table": "bloomerp_document_template_free_variable",
                "managed": True,
            },
        ),
        migrations.CreateModel(
            name="DocumentTemplateHeader",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                ("header", models.ImageField(upload_to="")),
            ],
            options={
                "db_table": "bloomerp_document_template_header",
                "managed": True,
            },
        ),
        migrations.RenameField(
            model_name="file",
            old_name="datetime_created",
            new_name="created_at",
        ),
        migrations.AddField(
            model_name="file",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="userkpidashboardviewpreference",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="userkpidashboardviewpreference",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="userkpilistviewpreference",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="userkpilistviewpreference",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="file",
            name="file",
            field=models.FileField(upload_to=bloomerp.models.File.upload_to),
        ),
        migrations.CreateModel(
            name="DocumentTemplate",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                ("template", models.TextField(default="")),
                ("footer", models.TextField(blank=True, null=True)),
                ("standard_document", models.BooleanField(default=False)),
                (
                    "model_variable",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "free_variables",
                    models.ManyToManyField(
                        blank=True, to="bloomerp.documenttemplatefreevariable"
                    ),
                ),
                (
                    "template_header",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="bloomerp.documenttemplateheader",
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_document_template",
                "managed": True,
            },
        ),
    ]



from django.conf import settings
from django import template

register = template.Library()

@register.simple_tag
def globals(name:str):
    '''
    Returns a global variable from the BLOOMERP_SETTINGS configuration in settings.py.

    Usage:
    {% load globals %}
    {% globals "variable_name" %}

    '''
    if not hasattr(settings, "BLOOMERP_SETTINGS"):
        raise ValueError("BLOOMERP_SETTINGS not found in settings.py. Please configure it according to the documentation.")
    
    globals:dict = settings.BLOOMERP_SETTINGS.get("globals")
    value = globals.get(name)
    return value




from setuptools import find_packages, setup


setup(
    name="Bloomerp",
    version="0.3.1",
    packages=find_packages(),
    include_package_data=True,
    license="MIT License",
    description="Bloomerp is an open-source Business Management Software framework that lets you create fully functional business management applications just by defining your Django database models.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/davidbloomer11/Bloomerp",
    website="https://bloomerp.io",
    author="David Bloomer",
    author_email="bloomer.david@outlook.com",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'requests>=2.0.0',
        'djangorestframework>=3.12.0',
        'django-htmx>=1.0.0',
        'django-formtools>=2.3',
        'Pillow>=8.0.0',
        'psycopg2==2.9.9',
        'django-crispy-forms>=2.2',
        'django-filter>=24.3',
        'crispy-bootstrap5==2024.2',
        'xhtml2pdf>=0.2.16',
        'pandas>=2.0.0',
        'openpyxl>=3.1.5',
        'openai>=1.54.1',
        'plotly>=5.0.0',
        'PyPDF2>=3.0.1',
        'langchain-core>=0.3.23',
        'langchain-openai>=0.2.12',
        'langgraph>=0.2.60',
        'colorama>=0.4.6',
    ],
    extras_require={
        'images': ['Pillow>=8.0.0'],
    },
)



from django.test import TestCase
from bloomerp.forms.auth import UserDetailViewPreferenceForm

class BloomerpFormsTests(TestCase):
    def test_user_detail_view_preference_form(self):
        form_data = {'field1': 'value1', 'field2': 'value2'}
        form = UserDetailViewPreferenceForm(data=form_data)
        self.assertTrue(form.is_valid())


from crispy_forms.helper import FormHelper
from bloomerp.models import BloomerpModel
from crispy_forms.layout import Layout, Fieldset, Submit, Div, HTML
from uuid import uuid4

class BloomerpModelformHelper(FormHelper):
    layout_defined: bool = False

    def __init__(self, model: BloomerpModel, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.form_tag = False

        if not model:
            self.layout_defined = False
            return

        content = '''<h3 class="dropdown-toggle form-legend pointer" onclick="document.getElementById('{id}').classList.toggle('d-none')">{title}{asterix}</h3>'''

        fieldsets = []

        if hasattr(model, 'form_layout') and model.form_layout:
            layout = model._get_form_layout()
            if not layout:
                self.layout_defined = False
                return


            for title, item in layout.items():
                id = uuid4()

                field_list = item['fields']
                required = item['required']

                css_class = '' if required else ''
                asterix = '*' if required else ''

                _content = content.format(title=title, id=id, asterix=asterix)

                fieldsets.append(
                    Div(
                        HTML(_content),
                        Fieldset(None, *field_list, css_id=id, css_class=css_class)
                    )
                )

            self.layout = Layout(
                *fieldsets
            )

            self.layout_defined = True
        else:
            self.layout_defined = False

    def is_defined(self) -> bool:
        return self.layout_defined




# Generated by Django 5.1.1 on 2024-10-14 19:53

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0035_remove_userappdashboard_layout_and_more"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="link",
            name="level",
            field=models.CharField(
                choices=[("DETAIL", "Detail"), ("LIST", "List"), ("APP", "App")],
                help_text="The level of the link.",
                max_length=255,
            ),
        ),
        migrations.CreateModel(
            name="UserDetailViewTabs",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "links",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The links to be displayed on the detail view tabs",
                        to="bloomerp.link",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django import forms
from django.forms import ModelChoiceField
from bloomerp.models.widgets import Widget, SqlQuery
from django.core.exceptions import ValidationError
from bloomerp.widgets.foreign_key_widget import ForeignKeyWidget
from bloomerp.widgets.text_editor import RichTextEditorWidget
from django.utils.translation import gettext as _

class WidgetForm1(forms.Form):
    query = ModelChoiceField(
        queryset=SqlQuery.objects.all(),
        widget=ForeignKeyWidget(model=SqlQuery),
        help_text=_('The sql query (data) to use for the query')
        
        )
    name = forms.CharField(
        max_length=255,
        help_text=_('The name given to the widget')
        )
    description = forms.CharField(
        widget=RichTextEditorWidget(),
        required=False,
        help_text=_('The description of the widget. Will be displayed in the widget if provided')
        )
    
    def __init__(self, widget:Widget=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if widget:
            self.fields['name'].initial = widget.name
            self.fields['description'].initial = widget.description
            self.fields['query'].initial = widget.query
        

class WidgetForm2(forms.Form):
    output_type = forms.ChoiceField(
        choices=Widget.OUTPUT_TYPE_CHOICES,
        help_text=_('The type of output to display for the widget')
        )
    
    def __init__(self, widget:Widget=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if widget:
            self.fields['output_type'].initial = widget.output_type

class WidgetForm3(forms.Form):
    def __init__(self, widget:Widget=None, *args, **kwargs):
        output_type = kwargs.pop('output_type', None)
        query : SqlQuery = kwargs.pop('query', None)

        super().__init__(*args, **kwargs)
        
        if query and output_type:
            # If the output type is table, add a field for the columns
            columns = query.result_dict()['columns']

            choices = [(None, '---------')] + [(col, col) for col in columns]

            if output_type == 'table':
                self.fields['columns'] = forms.MultipleChoiceField(
                    choices=[(col, col) for col in columns]
                    )
                self.fields['limit'] = forms.IntegerField(min_value=1, required=False, help_text=_('The number of rows to display in the table'), initial=10)
            elif output_type == 'scatter' or output_type == 'line' or output_type == 'bar':
                self.fields['x'] = forms.ChoiceField(choices=choices)
                self.fields['y'] = forms.ChoiceField(choices=choices)
                self.fields['group_by'] = forms.ChoiceField(choices=choices, required=False)
            elif output_type == 'pie':
                self.fields['x'] = forms.ChoiceField(choices=choices)
                self.fields['group_by'] = forms.ChoiceField(choices=choices)
            elif output_type == 'histogram':
                self.fields['x'] = forms.ChoiceField(choices=choices)
            elif output_type == 'value':
                self.fields['color'] = forms.CharField(widget=forms.TextInput(attrs={'type': 'color'}))
                self.fields['icon'] = forms.CharField(required=False, initial='bi bi-bar-chart-fill')
                self.fields['aggregate'] = forms.ChoiceField(choices=Widget.VALUE_AGGREGATE_CHOICES)
                self.fields['column'] = forms.ChoiceField(choices=choices, required=True)
                self.fields['prefix'] = forms.CharField(max_length=255, required=False)
                self.fields['suffix'] = forms.CharField(max_length=255, required=False)
        
    def clean(self):
        cleaned_data = super().clean()
        output_type = cleaned_data.get('output_type')
        columns = cleaned_data.get('columns')
        x = cleaned_data.get('x')
        y = cleaned_data.get('y')
        group_by = cleaned_data.get('group_by')
        color = cleaned_data.get('color')

        if output_type == 'table' and not columns:
            raise ValidationError('Columns must be selected for table output type')
        elif output_type in ['scatter', 'line', 'bar']:
            if not x:
                raise ValidationError('X-axis must be selected for scatter, line or bar output type')
            if not y:
                raise ValidationError('Y-axis must be selected for scatter, line or bar output type')              

        elif output_type == 'pie':
            if not x:
                raise ValidationError('X-axis must be selected for pie output type')
            if not group_by:
                raise ValidationError('Group by must be selected for pie output type')
        elif output_type == 'histogram':
            if not x:
                raise ValidationError('X-axis must be selected for histogram output type')
        elif output_type == 'value':
            if not color:
                raise ValidationError('Color must be selected for value output type')

        return cleaned_data



class SqlQueryForm(forms.Form):
    code = forms.CharField(widget=forms.HiddenInput(), required=False)  # Hidden field to store code content



from typing import Any
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseRedirect,
)
from django.shortcuts import render
from django.urls import reverse
from django.views.generic.detail import DetailView
from django.views.generic.edit import UpdateView, CreateView
from django.views.generic.list import ListView
from django.views.generic import View
from django.contrib.auth.mixins import PermissionRequiredMixin, LoginRequiredMixin
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import (
    ApplicationField,
    File,
    Widget,
    SqlQuery,
    UserDetailViewPreference,
    Bookmark,
    User
)
from bloomerp.forms.core import BloomerpDownloadBulkUploadTemplateForm
from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from bloomerp.utils.models import model_name_plural_underline, get_list_view_url, get_detail_view_url
from django.db.models import Model
from bloomerp.views.mixins import (
    BloomerpModelFormViewMixin,
    BloomerpModelContextMixin,
    HtmxMixin
)

from bloomerp.utils.router import BloomerpRouter, _get_name_or_slug
from django.forms.models import modelform_factory
from bloomerp.forms.core import BloomerpModelForm
from django import forms

router = BloomerpRouter()

# ---------------------------------
# Bloomerp List View
# ---------------------------------
@router.bloomerp_route(
    path="list",
    name="{model} list",
    url_name="list",
    description="List of {model} model",
    route_type="list",
    exclude_models=[File]
)
class BloomerpListView(
        PermissionRequiredMixin,
        BloomerpModelContextMixin, 
        HtmxMixin, 
        ListView):
    model : Model = None # The model to use for the view
    template_name : str = "list_views/bloomerp_list_view.html" # The template to use for rendering the view
    context_object_name : str = "object_list" # The name of the context variable to use in the template
    create_object_url : str = None
    permission_required = None

    def get_permission_required(self):
        if self.permission_required:
            return self.permission_required
        else:
            return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]
        
    def get_context_data(self, **kwargs: Any) -> dict:
        context = super().get_context_data(**kwargs)
        context["title"] = self.model._meta.verbose_name.capitalize() + " list"
        return context

    
# ---------------------------------
# Bloomerp Detail View
# ---------------------------------
class BloomerpBaseDetailView(HtmxMixin, BloomerpModelContextMixin, DetailView):
    htmx_template = 'bloomerp_htmx_base_view.html'
    tabs = None
    exclude_header = False

    def get_context_data(self, **kwargs: Any) -> dict:
        context = super().get_context_data(**kwargs)
        context["exclude_header"] = self.exclude_header

        if self.tabs:
            context["tabs"] = self.tabs

        return context

# ---------------------------------
# Bloomerp Detail Overview View
# ---------------------------------
@router.bloomerp_route(
    path="",
    name="Overview",
    url_name="overview",
    description="Overview of object from {model} model",
    route_type="detail",
    models = "__all__",
)
class BloomerpDetailOverviewView(PermissionRequiredMixin, BloomerpBaseDetailView):
    template_name = "detail_views/bloomerp_detail_overview_view.html"
    settings = None
    
    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]

    def get_context_data(self, **kwargs: Any) -> dict:
        context = super().get_context_data(**kwargs)

        content_type = ContentType.objects.get_for_model(self.model)

        queryset = UserDetailViewPreference.objects.filter(
            user=self.request.user, application_field__content_type=content_type
        )

        # Add content type id to context
        context["view_type"] = "DETAIL"
        context["content_type_id"] = content_type.pk
        if not queryset:
            queryset = UserDetailViewPreference.generate_default_for_user(
                user = self.request.user,
                content_type = content_type
            )
            
        left_column = queryset.filter(position="LEFT")
        center_column = queryset.filter(position="CENTER")
        right_column = queryset.filter(position="RIGHT")

        # Do some processing
        i = 0
        if left_column:
            i += 1
        if center_column:
            i += 1
        if right_column:
            i += 1

        if i == 0:
            col_span = 12
        else:
            col_span = 12 / i

        context["col_span"] = int(col_span)
        context["left_column"] = left_column
        context["right_column"] = right_column
        context["center_column"] = center_column

        return context

# ---------------------------------
# Bloomerp Create View
# ---------------------------------
@router.bloomerp_route(
    path="create",
    name="Create {model}",
    url_name="add",
    description="Create a new object from {model}",
    route_type="list",
    exclude_models=[File, Widget, SqlQuery, User]
)
class BloomerpCreateView(
        PermissionRequiredMixin, 
        SuccessMessageMixin,
        HtmxMixin, 
        BloomerpModelFormViewMixin,
        CreateView):
    template_name = "create_views/bloomerp_create_view.html"
    fields = None
    exclude = []
    success_message = "Object was created successfully."

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["model_name"] = self.model._meta.verbose_name
        context["model_name_plural"] = self.model._meta.verbose_name_plural
        context["list_view_url"] = model_name_plural_underline(self.model) + "_list"
        context["model"] = self.model
        context["title"] = 'Create ' + self.model._meta.verbose_name
        return context

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.add_{self.model._meta.model_name}"]

    def get_success_message(self, cleaned_data):
        return f"{self.object} was created successfully."
    
    def get_success_url(self):
        try:
            return self.object.get_absolute_url()
        except AttributeError:
            return reverse(get_detail_view_url(self.object), kwargs={"pk": self.object.pk})

# ---------------------------------
# Bloomerp Update View
# ---------------------------------
@router.bloomerp_route(
    path="update",
    name="Update",
    url_name="update",
    description="Update object from {model}",
    route_type="detail",
    exclude_models=[File, Widget]
)
class BloomerpUpdateView(
        PermissionRequiredMixin, 
        SuccessMessageMixin, 
        HtmxMixin,
        BloomerpModelFormViewMixin,
        BloomerpModelContextMixin,
        UpdateView
        ):
    template_name = "detail_views/bloomerp_detail_update_view.html"
    settings = None
    _uses_base_form = False

    def get_success_url(self):
        try:
            return self.object.get_absolute_url()
        except AttributeError:
            return reverse(get_detail_view_url(self.object), kwargs={"pk": self.object.pk})


    def get_permissions(self):
        """
        Returns a dictionary of standard add, change, delete, and view permissions
        for the given model.

        Returns:
        Dictionary with keys 'add', 'change', 'delete', 'view' and their corresponding
        permission codenames as values.
        """
        model_name = self.model._meta.model_name  # Get the model name in lowercase
        app_label = self.model._meta.app_label  # Get the app label

        permissions = {
            "create": f"{app_label}.add_{model_name}",
            "update": f"{app_label}.change_{model_name}",
            "delete": f"{app_label}.delete_{model_name}",
            "read": f"{app_label}.view_{model_name}",
        }

        return permissions

    def get_permission_required(self):
        return [self.get_permissions()["update"]]

    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        if request.POST.get("delete"):
            if not request.user.has_perm(self.get_permissions()["delete"]):
                messages.error(
                    request,
                    f'User does not have the required permission: {self.get_permissions()["delete"]}.',
                )
                return HttpResponseRedirect(request.path)

            self.get_object().delete()
            messages.info(request, f"Object was deleted successfully.")
            
            url = get_list_view_url(self.model)

            return HttpResponseRedirect(reverse(url))

        return super().post(request, *args, **kwargs)

    def get_success_message(self, cleaned_data):
        return f"{self.object} was updated successfully."

# ---------------------------------
# Bloomerp many-to-many detail view
# ---------------------------------
from django.db.models.fields.reverse_related import ManyToManyRel, ManyToOneRel
from bloomerp.models import Comment
def get_view_parameters(model:Model):
    params_list = []
    model_name = model._meta.verbose_name

    fields = model._meta.get_fields()

    if model in [File, Comment]:
        return False

    for field in fields:
        # Skip fields that 
        if field.name in ['created_by','updated_by']:
            continue

        if field.get_internal_type() == 'ManyToManyField':
            if field.related_model in [Comment, File]:
                continue

            if type(field) == ManyToManyRel:
                params_dict = {
                    'path' : f'{_get_name_or_slug(field.related_model, slug=True)}',
                    'name' : f'{field.related_model._meta.verbose_name_plural.capitalize()}',
                    'url_name' : f'{_get_name_or_slug(field.related_model)}_relationship',
                    'model' : model,
                    'route_type' : 'detail',
                    'description' : f'{field.related_model._meta.verbose_name_plural.capitalize()} relationship for {model_name}',
                    'args': {
                        'related_model':field.related_model,
                        'related_model_attribute' : field.name,
                        'reversed':True,
                        'related_model_field' : field.remote_field.name
                    }
                }
                
            else:
                if field.related_model in [Comment, File]:
                    continue

                params_dict = {
                    'path' : f'{field.name}/',
                    'name' : f'{field.verbose_name.capitalize()}',
                    'url_name' : field.name + '_relationship',
                    'description' : f'{field.verbose_name.capitalize()} relationship for {model_name}',
                    'model' : model,
                    'route_type' : 'detail',
                    'args' : {
                        'related_model':field.related_model,
                        'related_model_attribute':True,
                        'reversed':False,
                        'related_model_field':field.remote_field.name
                        }
                }

            params_list.append(params_dict)

        elif type(field) == ManyToOneRel:
            if field.related_model in [Comment, File]:
                return False

            params_dict = {
                'path' : f'{_get_name_or_slug(field.related_model, slug=True)}',
                'name' : f'{field.name.capitalize().replace('_',' ')}',
                'url_name' : f'{_get_name_or_slug(field.related_model)}_relationship',
                'model' : model,
                'route_type' : 'detail',
                'description' : f'{field.name.capitalize().replace('_',' ')} relationship for {model_name}',
                'args': {
                    'related_model':field.related_model,
                    'related_model_attribute' : field.name,
                    'reversed':False,
                    'related_model_field' : field.remote_field.name
                }
            }
        
            params_list.append(params_dict)

    return params_list


@router.bloomerp_route(from_func=get_view_parameters)
class BloomerpDetailM2MView(
        PermissionRequiredMixin,
        BloomerpBaseDetailView
        ):
    template_name : str = "detail_views/bloomerp_detail_m2m_view.html"
    model : Model = None
    related_model : Model = None
    related_model_attribute : str = None
    reversed : bool = None
    related_model_field : str = None

    def get_permission_required(self):
        foreign_view_permission = f"{self.related_model._meta.app_label}.view_{self.related_model._meta.model_name}"
        model_view_permission = (
            f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"
        )
        return [foreign_view_permission, model_view_permission]

    def get_context_data(self, **kwargs: Any) -> dict:
        context = super().get_context_data(**kwargs)
        
        # Construct initial query
        initial_query = f'{self.related_model_field}={self.get_object().pk}'        

        # Get application fields for the foreign model
        application_fields = ApplicationField.get_for_model(self.related_model)
        if self.reversed == False:
            # Create form
            Form = modelform_factory(model=self.related_model, fields='__all__', form=BloomerpModelForm)
            
            form = Form(model=self.related_model, user=self.request.user, initial={self.related_model_field:self.get_object().pk})

            context['form'] = form

        # Set content_type_id 
        context['foreign_content_type_id'] = ContentType.objects.get_for_model(self.related_model).pk
        context['foreign_model_attribute'] = self.related_model_attribute
        context['object'] = self.object
        context['application_fields'] = application_fields
        context['initial_query'] = initial_query
        context['reversed'] = self.reversed
        context['related_model'] = self.related_model
        context['related_model_name_singular'] = self.related_model._meta.verbose_name
        context['related_model_name_plural'] = self.related_model._meta.verbose_name_plural
        return context


# ---------------------------------
# Bloomerp Detail File View
# ---------------------------------
@router.bloomerp_route(
    path="files",
    name="Files",
    url_name="files",
    description="Files for object for {model} model",
    route_type="detail",
    exclude_models=[File]
)
class BloomerpDetailFileListView(PermissionRequiredMixin, BloomerpBaseDetailView):
    template_name = "snippets/files_snippet.html"
    
    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        fields = ['name', 'datetime_created', 'datetime_updated', 'created_by','updated_by', 'object_id', 'content_type']
        
        context['files_application_fields'] = ApplicationField.objects.filter(content_type=ContentType.objects.get_for_model(File), field__in=fields)
        context['title'] = f'Files for {self.get_object()}'
        context['target'] = 'file_results'
        return context

# ---------------------------------
# Bloomerp Bulk Upload View
# ---------------------------------
@router.bloomerp_route(
    path="bulk-upload",
    name="Bulk Upload {model}",
    url_name="bulk_upload",
    description="Bulk upload objects from {model}",
    route_type="list",
    exclude_models=[File]
)
class BloomerpBulkUploadView(PermissionRequiredMixin, HtmxMixin, View):
    template_name = "list_views/bloomerp_bulk_upload_view.html"
    model = None
    success_url = None
    success_message = "Objects were uploaded successfully."
    
    def get_permission_required(self):
        # Probably want to add a bulk upload permission here
        return [f"{self.model._meta.app_label}.bulk_add_{self.model._meta.model_name}"]

    def get_context_data(self, **kwargs) -> dict:
        context = super().get_context_data(**kwargs)
        context["model_name"] = self.model._meta.verbose_name
        context["model"] = self.model
        context["title"] = f'Bulk upload {self.model._meta.verbose_name_plural}'
        context["model_name_plural"] = self.model._meta.verbose_name_plural
        context["content_type_id"] = ContentType.objects.get_for_model(self.model).pk
        context["list_view_url"] = model_name_plural_underline(self.model) + "_list"
        context["fields_form"] = BloomerpDownloadBulkUploadTemplateForm(self.model)
        return context

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        return render(request, self.template_name, context)


# ---------------------------------
# Bloomerp File List View
# ---------------------------------
@router.bloomerp_route(
    path="list",
    name="Files",
    url_name="list",
    description="Bloomerp Files",
    route_type="list",
    models = [File]
)
class BloomerpFileListView(PermissionRequiredMixin, HtmxMixin, View):
    template_name = "list_views/bloomerp_file_list_view.html"
    model = File

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Get the application fields for filter purposes
        fields = ['name', 'datetime_created', 'datetime_updated', 'created_by','updated_by', 'object_id', 'content_type']
        context['target'] = 'file_list'
        context['application_fields'] = ApplicationField.objects.filter(content_type=ContentType.objects.get_for_model(self.model), field__in=fields)
        return context

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]
    
    def get(self, request, *args, **kwargs):
        context = self.get_context_data()

        return render(request, self.template_name, context)
    
# ---------------------------------
# Bloomerp Detail Comments View
# ---------------------------------
from bloomerp.models import Comment
@router.bloomerp_route(
    path="comments",
    name="Comments",
    url_name="{model}_detail_comments",
    description="Comments for object from {model} model",
    route_type="detail",
    models="__all__"
)
class BloomerpDetailCommentsView(PermissionRequiredMixin, BloomerpBaseDetailView):
    template_name = "detail_views/bloomerp_detail_comments_view.html"
    model = None

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if not hasattr(self.object, 'comments'):
            comments = Comment.objects.filter(content_type=self.content_type, object_id=self.object.pk)
        else:
            comments = self.object.comments.all()

        context['comments'] = comments
        return context

# ---------------------------------
# Bloomerp Bookmarks View
# ---------------------------------
@router.bloomerp_route(
    path="user-bookmarks",
    name="Bookmarks for user",
    url_name="user_bookmarks",
    description="Bookmarks for user",
    route_type="list",
    models=[Bookmark]
)
class BloomerpBookmarksView(LoginRequiredMixin, HtmxMixin, View):
    template_name = "list_views/bloomerp_bookmarks_view.html"
    model = None

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['content_type_id'] = ContentType.objects.get_for_model(self.model).pk
        context['initial_query'] = 'user=' + str(self.request.user.pk)
        return context

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        return render(request, self.template_name, context)




'''
from formtools.wizard.views import SessionWizardView
from collections import OrderedDict
from django import forms
from modules.documents.forms import generate_fields_for_model_document_templates
from django.core.files.storage import FileSystemStorage
import os
from django.conf import settings
class BaseWizzardCreateView(
    PermissionRequiredMixin, SuccessMessageMixin, SessionWizardView
):
    template_name = "bloomerp_base_wizard_create_view.html"
    model = None
    fields = None
    exclude = []
    success_message = "Object was created successfully."
    form_list = [BaseProjectModelForm]
    permission_required = None
    file_storage = FileSystemStorage(
        location=os.path.join(settings.MEDIA_ROOT, "temporary")
    )
    model_form = None
    document_generator = True

    def get_permission_required(self):
        if self.permission_required:
            return self.permission_required
        else:
            return [f"{self.model._meta.app_label}.add_{self.model._meta.model_name}"]

    def get_context_data(self, form, **kwargs):
        context = super().get_context_data(form, **kwargs)
        context["model_name"] = self.model._meta.verbose_name
        return context

    def get_form_kwargs(self, step=None):
        kwargs = super().get_form_kwargs(step)

        if step == "last":
            return kwargs

        kwargs["user"] = self.request.user

        return kwargs

    def get_form_list(self):
        # Create a form list with the model form
        form_list = OrderedDict()

        # Check if a model form is provided
        if self.model_form:
            Form = self.model_form
        else: 
            Form = modelform_factory(
                model=self.model,
                fields="__all__",
                form=BaseProjectModelForm,
                exclude=self.exclude + ["last_updated_by"],
            )

        form_list["0"] = None

        # Check if model already has any items in the database to avoid RelatedObjectDoesNotExist error, if not, skip the foreign relationships
        # This error only occurs when a foreing relationship is mandatory
        # This is a workaround to avoid the error
        has_items = self.model.objects.first()

        # Get related models
        qs = ApplicationField.get_related_models(self.model, skip_auto_created=False)
        qs = qs.exclude(field_type="OneToManyField").exclude(
            field_type="ManyToManyField"
        )

        self.application_fields = qs

        for i, application_field in enumerate(qs):
            # Get the model from the application field
            model = application_field.content_type.model_class()

            # Check if model has any items in the database
            if model.objects.exists():
                ForeignForm = modelform_factory(
                    model=model, fields="__all__", form=BaseProjectModelForm
                )
                # Hide the foreign key field

                ForeignForm.base_fields[
                    application_field.field
                ].widget = forms.HiddenInput()

                form_list[str(i + 1)] = ForeignForm

                Form.base_fields[f"add_{model._meta.model_name}"] = forms.BooleanField(
                    label=f"Add {model._meta.verbose_name}", required=False
                )

                # Create lambda function to set conditional form
                conditional_function = lambda wizard, model=model: (
                    wizard.get_cleaned_data_for_step("0") or {}
                ).get(f"add_{model._meta.model_name}")
                self.condition_dict[str(i + 1)] = conditional_function

        # Add conditional form to generate documents
        Form.base_fields["generate_documents"] = forms.BooleanField(
            label="Generate documents", required=False
        )
        self.condition_dict["last"] = lambda wizard: (
            wizard.get_cleaned_data_for_step("0") or {}
        ).get("generate_documents")
        fields = generate_fields_for_model_document_templates(
            self.model, only_standard_documents=True
        )

        if self.document_generator:
            DocumentTemplateForm = forms.Form
            for field in fields:
                DocumentTemplateForm.base_fields[field[0]] = field[1]

            form_list["last"] = DocumentTemplateForm
        form_list["0"] = Form

        self.form_list = form_list

        return super().get_form_list()

    def get_form_initial(self, step):
        """Sets the initial foreign relationships"""
        initial = super().get_form_initial(step)

        for i, application_field in enumerate(self.application_fields):
            # Set the initial value for the foreign relationship form to a random object
            random_object = self.model.objects.first()
            initial[application_field.field] = random_object

        return initial

    def done(self, form_list, form_dict, **kwargs):
        # Create the main object
        main_obj = form_dict["0"].save(commit=False)

        # Set the last_updated_by attribute if it exists
        if hasattr(main_obj, "last_updated_by"):
            main_obj.last_updated_by = self.request.user

        # Save the main object
        main_obj.save()

        # Save many-to-many relationships if needed
        if hasattr(form_dict["0"], "save_m2m"):
            form_dict["0"].save_m2m()

        if hasattr(form_dict["0"], "update_file_fields"):
            form_dict["0"].update_file_fields()

        # Create the foreign relationships
        for i, application_field in enumerate(self.application_fields):
            if form_dict.get(str(i + 1)):
                foreign_obj = form_dict[str(i + 1)].save(commit=False)
                setattr(foreign_obj, application_field.field, main_obj)
                foreign_obj.save()

        # Generate documents from document templates
        templates = DocumentTemplate.get_standard_documents_for_instance(main_obj)
        if form_dict.get("last"):
            free_variables = form_dict["last"].cleaned_data
            controller = DocumentController()

            for template in templates:
                try:
                    controller.create_document(template, main_obj, free_variables)
                except Exception as e:
                    messages.error(
                        self.request, f"Error generating document {template.name}: {e}"
                    )

        return redirect(main_obj.get_absolute_url())

    def get_success_message(self, cleaned_data):
        return f"{self.object} was created successfully."

'''


from django.shortcuts import render
from bloomerp.models import Widget
from django.views import View
from django.contrib.auth.mixins import LoginRequiredMixin
from django.db.models import Model
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import Link, Workspace
from bloomerp.utils.router import BloomerpRouter
from bloomerp.views.mixins import HtmxMixin
from django.shortcuts import get_object_or_404

router = BloomerpRouter()	

@router.bloomerp_route(
    path='',
    name='Bloomerp Dashboard',
    description='The dashboard for the Bloomerp app',
    route_type='app',
    url_name='bloomerp_home_view'
)
class BloomerpHomeView(HtmxMixin, LoginRequiredMixin, View):
    template_name = 'workspace_views/bloomerp_workspace_view.html'

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        widget_list = Widget.objects.all()
        context['widget_list'] = widget_list

        # Create a workspace for the user if it doesn't exist
        workspace = Workspace.objects.filter(user=request.user, content_type=None).first()

        if not workspace:
            workspace = Workspace.create_default_workspace(request.user)
        
        context['workspace'] = workspace
        
        return render(request, self.template_name, context)

@router.bloomerp_route(
    path='workspace/<int:workspace_id>/',
    name='View Workspace',
    description='View a workspace',
    route_type='app',
    url_name='view_workspace'
)
class WorkspaceView(HtmxMixin, LoginRequiredMixin, View):
    template_name = 'workspace_views/bloomerp_workspace_view.html'

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        workspace_id = kwargs.get('workspace_id')
        workspace = get_object_or_404(Workspace, id=workspace_id)
        context['workspace'] = workspace
        return render(request, self.template_name, context)
    

@router.bloomerp_route(
        models="__all__",
        path='',
        name='{model} Dashboard',
        description='The dashboard for the {model} model',
        route_type = 'list',
        url_name='dashboard'
)
class BloomerpContentTypeWorkspaceView(
        HtmxMixin,
        LoginRequiredMixin,
        View
    ):
    model : Model = None
    template_name = 'workspace_views/bloomerp_workspace_view.html'

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        
        # Get the content type for the model
        content_type = ContentType.objects.get_for_model(self.model)
        
        workspace = Workspace.objects.filter(user=request.user, content_type=content_type).first()

        if not workspace:
            workspace = Workspace.create_default_content_type_workspace(request.user, content_type)

        context['workspace'] = workspace
        context['model_name_plural'] = self.model._meta.verbose_name_plural
        return render(request, self.template_name, context)


# Generated by Django 5.1.1 on 2025-01-04 12:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0060_documenttemplate_save_to_folder"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="documenttemplate",
            name="avatar",
        ),
        migrations.RemoveField(
            model_name="documenttemplatefreevariable",
            name="avatar",
        ),
        migrations.RemoveField(
            model_name="documenttemplateheader",
            name="avatar",
        ),
        migrations.RemoveField(
            model_name="documenttemplatestyling",
            name="avatar",
        ),
        migrations.RemoveField(
            model_name="sqlqueryfilter",
            name="filter_type",
        ),
        migrations.RemoveField(
            model_name="sqlqueryfilter",
            name="query",
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="column",
            field=models.CharField(
                default="employee_id",
                help_text="Column name to filter on",
                max_length=255,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="data_type",
            field=models.CharField(
                choices=[
                    ("STRING", "String"),
                    ("INTEGER", "Integer"),
                    ("FLOAT", "Float"),
                    ("DATE", "Date"),
                    ("DATETIME", "Datetime"),
                    ("BOOLEAN", "Boolean"),
                ],
                default="STRING",
                help_text="Data type of the filter",
                max_length=255,
            ),
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="filter_operation",
            field=models.CharField(
                choices=[
                    ("=", "Equal"),
                    (">", "Greater"),
                    ("<", "Less"),
                    (">=", "Greater Or Equal"),
                    ("<=", "Less Or Equal"),
                    ("!=", "Not Equal"),
                    ("LIKE", "Like"),
                    ("NOT LIKE", "Not Like"),
                    ("IN", "In"),
                    ("NOT IN", "Not In"),
                    ("IS", "Is"),
                    ("IS NOT", "Is Not"),
                    ("BETWEEN", "Between"),
                    ("NOT BETWEEN", "Not Between"),
                    ("REGEXP", "Regexp"),
                    ("NOT REGEXP", "Not Regexp"),
                    ("EXISTS", "Exists"),
                    ("NOT EXISTS", "Not Exists"),
                ],
                default="=",
                help_text="Type of the filter",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="sqlqueryfilter",
            name="name",
            field=models.CharField(help_text="Name of the filter", max_length=255),
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-16 18:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0058_alter_documenttemplate_model_variable"),
    ]

    operations = [
        migrations.AddField(
            model_name="documenttemplate",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="documenttemplatefreevariable",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="documenttemplatestyling",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="sqlquery",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="sqlqueryfilter",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="widget",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
    ]



from django import forms
from django.forms.models import modelform_factory
from django.db.models import Model
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import ApplicationField
import random
class MultipleModelSelect(forms.SelectMultiple):
    template_name = 'widgets/foreign_field_widget.html'

    def __init__(self, model: Model, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)
    
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        # Create new form in case user wants to update the foreign key
        Form = modelform_factory(self.model, fields='__all__')
        
        # Get first 5 objects from the model
        context['objects'] = self.model.objects.all()[:5]
        context['content_type_id'] = ContentType.objects.get_for_model(self.model).id
        context['form'] = Form()

        # Add random modal id to the context
        context['advanced_search_modal_id'] = f'modal-{random.randint(0, 100000)}'
        context['create_modal_id'] = f'modal-{random.randint(0, 100000)}'

        # Get the object(s) that is currently selected
        if context['widget']['value']:
            values : str = context['widget']['value']
            selected_choices = self.model.objects.filter(pk__in=values)
            context['selected_choices'] = selected_choices

        # Check if an invalid entry was made
        if attrs.get('aria-invalid', 'false') == 'true':
            context['invalid'] = True

        # Add application fields to the context
        context['application_fields'] = ApplicationField.objects.filter(content_type=ContentType.objects.get_for_model(self.model)).exclude(field_type='Property')

        context['widget_type'] = 'm2m'
        return context
    
        
        





from django.db import models
from django.forms import ValidationError
from bloomerp.models.core import BloomerpModel, File
from bloomerp.models.fields import BloomerpFileField, StatusField
from django.utils.translation import gettext as _

# Create your models here.
class Customer(BloomerpModel):
    name = models.CharField(max_length=255, help_text="The name of the customer")
    email = models.EmailField()
    phone = models.CharField(max_length=20)
    address = models.TextField()
    city = models.CharField(max_length=100)
    state = models.CharField(max_length=100)
    zip_code = models.CharField(max_length=10)
    country = models.CharField(max_length=100)
    notes = models.TextField(blank=True, null=True)

    string_search_fields = ['name']
    allow_string_search = True

    def __str__(self):
        return self.name

class Product(BloomerpModel):
    name = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)
    price = models.DecimalField(max_digits=10, decimal_places=2)
    stock = models.IntegerField(default=0)
    image = models.ImageField(upload_to="products", blank=True, null=True)
    
    string_search_fields = ['name']
    allow_string_search = True

    def __str__(self):
        return self.name
    
class Order(BloomerpModel):
    customer = models.ForeignKey(Customer, on_delete=models.CASCADE, related_name="orders")
    products = models.ManyToManyField(Product)
    total = models.DecimalField(max_digits=10, decimal_places=2)
    status = StatusField(
        colored_choices = [
            ('pending', 'Pending', '#ffcc00'),
            ('processing', 'Processing', '#007bff'),
            ('completed', 'Completed', '#28a745'),
            ('cancelled', 'Cancelled', '#dc3545'),
        ],
        max_length=30
    )
    notes = models.TextField(blank=True, null=True)
    date = models.DateTimeField(blank=True, null=True)


    string_search_fields = ['customer__name', 'status', 'notes', 'products__name']

    def __str__(self):
        return f"Order {self.id} by {self.customer.name}"

class Function(BloomerpModel):
    name = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)
    
    allow_string_search = True
    string_search_fields = ['name']

    def __str__(self):
        return self.name

class Equipment(BloomerpModel):
    id = models.UUIDField(primary_key=True)
    name = models.CharField(max_length=255)

class Employee(BloomerpModel):
    first_name = models.CharField(max_length=255, help_text=_("The first name of the employee"))
    last_name = models.CharField(max_length=255, help_text=_("The last name of the employee"))
    email = models.EmailField()
    identification_document = BloomerpFileField( null=True, blank=True, allowed_extensions=['.jpg', '.png', '.jpeg'])
    manager = models.ForeignKey('self', on_delete=models.CASCADE, null=True, blank=True, help_text=_("The manager of the employee"), related_name="direct_reports")
    functions = models.ManyToManyField(Function)
    equipment = models.ManyToManyField(Equipment)
    date_of_birth = models.DateField(blank=True, null=True)
    gender = models.CharField(max_length=1, choices=[
        ('M', 'Male'),
        ('F', 'Female'),
        ('O', 'Other'),
    ], blank=True, null=True)
    
    string_search_fields = ['first_name', 'last_name']
    allow_string_search = True

    form_layout = {
        "General Information" : [
            "first_name",
            "last_name",
            "email",
            "identification_document",
            "date_of_birth",
            "gender"
        ],
        "Manager Information" : [
            "manager",
        ],
        "Functions" : [
            "functions",
            "equipment",
        ],
    }


    def __str__(self):
        return self.first_name + " " + self.last_name

    def clean(self):
        errors = {}

        if self.manager == self:
            errors['manager'] = "Cannot be the same as the employee"
        

        if errors:
            raise ValidationError(errors)


class Accommodation(BloomerpModel):
    name = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)
    price = models.DecimalField(max_digits=10, decimal_places=2)
    image = BloomerpFileField(allowed_extensions=['.jpg', '.png', '.jpeg'], null=True, blank=True)
    
    string_search_fields = ['name']

    def __str__(self):
        return self.name
    

class EmployeeOnAccommodation(BloomerpModel):
    employee = models.ForeignKey(Employee, on_delete=models.CASCADE)
    accommodation = models.ForeignKey(Accommodation, on_delete=models.CASCADE)
    check_in = models.DateField()
    check_out = models.DateField()
    
    string_search_fields = ['employee__first_name', 'employee__last_name', 'accommodation__name']

    def __str__(self):
        return f"{self.employee} on {self.accommodation.name}"


    def clean(self):
        errors = {}

        if self.check_in >= self.check_out:
            errors['check_in'] = "Check in date must be before check out date"
            errors['check_out'] = "Check out date must be after check in date"
        
        if errors:
            raise ValidationError(errors)
        
        super().clean()
    


# Generated by Django 5.1.1 on 2025-01-11 00:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0068_aiconversation"),
    ]

    operations = [
        migrations.AlterField(
            model_name="aiconversation",
            name="conversation_history",
            field=models.JSONField(blank=True, null=True),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-26 23:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0012_equipment_employee_equipment"),
    ]

    operations = [
        migrations.AlterField(
            model_name="customer",
            name="name",
            field=models.CharField(
                help_text="The name of the customer", max_length=255
            ),
        ),
        migrations.AlterField(
            model_name="employee",
            name="first_name",
            field=models.CharField(
                help_text="The first name of the employee", max_length=255
            ),
        ),
        migrations.AlterField(
            model_name="employee",
            name="last_name",
            field=models.CharField(
                help_text="The last name of the employee", max_length=255
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-20 22:12

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0048_remove_userlistviewpreference_is_main"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="file",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="filefolder",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="widget",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-26 15:45

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0010_alter_customer_options"),
    ]

    operations = [
        migrations.CreateModel(
            name="Function",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(blank=True, null=True)),
            ],
            options={
                "permissions": (
                    (
                        "bulk_delete%(class)s",
                        "Can delete multiple objects from %(class)s at once",
                    ),
                    (
                        "bulk_change%(class)s",
                        "Can update multiple objects from %(class)s at once",
                    ),
                ),
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="employee",
            name="manager",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="test_app.employee",
            ),
        ),
        migrations.AddField(
            model_name="employee",
            name="functions",
            field=models.ManyToManyField(to="test_app.function"),
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-28 17:17

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0015_employee_date_of_birth"),
    ]

    operations = [
        migrations.AlterField(
            model_name="employee",
            name="manager",
            field=models.ForeignKey(
                blank=True,
                help_text="The manager of the employee",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="direct_reports",
                to="test_app.employee",
            ),
        ),
    ]



from django.contrib.auth.mixins import PermissionRequiredMixin
from formtools.wizard.views import SessionWizardView
from django.shortcuts import redirect
from bloomerp.forms.widgets import WidgetForm1, WidgetForm2, WidgetForm3
from bloomerp.models import Widget, SqlQuery, ApplicationField
from bloomerp.utils.router import BloomerpRouter
from bloomerp.views.mixins import HtmxMixin, BloomerpModelFormViewMixin, BloomerpModelContextMixin
from bloomerp.views.core import BloomerpBaseDetailView
from django.views.generic.edit import CreateView, UpdateView
from django.views.generic.detail import DetailView
from bloomerp.components.workspaces.workspace_widget import workspace_widget


router = BloomerpRouter()

# ---------------------------------
# Sql Query Create View
# ---------------------------------
@router.bloomerp_route(
    models = SqlQuery,
    path='create',
    name= 'Create Sql Query',
    description='Create a new Sql Query',
    route_type='list',
    url_name='add'
)
class SqlQueryCreateView(PermissionRequiredMixin, HtmxMixin, BloomerpModelFormViewMixin, CreateView):
    model = SqlQuery
    template_name = 'widget_views/base_sql_query_builder_view.html'
    permission_required = ['widget.add_widget']
    fields = ['name', 'query']
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['model'] = self.model
        context['db_tables_and_columns'] = ApplicationField.get_db_tables_and_columns()
        return context
    
# ---------------------------------
# Sql Query Editor View
# ---------------------------------
@router.bloomerp_route(
    models = SqlQuery,
    path='editor',
    name= 'Sql Query Editor',
    description='Edit a Sql Query',
    route_type='detail',
    url_name='editor'
)
class SqlQueryEditorView(PermissionRequiredMixin, BloomerpModelContextMixin, HtmxMixin, BloomerpModelFormViewMixin, UpdateView):
    model = SqlQuery
    template_name = "widget_views/snippets/sql_query_builder_snippet.html"
    permission_required = ['bloomerp.execute_sql_query']
    fields = ['name', 'query']
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['model'] = self.model
        context['db_tables_and_columns'] = ApplicationField.get_db_tables_and_columns()
        return context

    def get_object(self, queryset=None):
        return self.model.objects.get(pk=self.kwargs.get('pk'))

    def get_success_url(self):
        return self.object.get_absolute_url()

    





# ---------------------------------
# Widget Create Wizard View
# ---------------------------------
@router.bloomerp_route(
        models=Widget,
        path='create',
        name='Create Widget',
        description='Create a new Widget',
        route_type='list',
        url_name='add'
)
class WidgetCreateWizardView(PermissionRequiredMixin, HtmxMixin, SessionWizardView):
    form_list = [WidgetForm1, WidgetForm2, WidgetForm3]
    template_name = 'create_views/bloomerp_base_wizard_create_view.html'
    permission_required = 'shared_utils.add_widget'
    model = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['model_name_plural'] = 'Widgets'
        context['model_name'] = 'Widget'
        context['model'] = self.model
        return context

    def get_form_kwargs(self, step=None):
        """
        Returns the keyword arguments for instantiating the form
        (or formset) on the given step.
        """
        if step == '2':
            # Fetch the output type selected in the previous step
            output_type = self.get_cleaned_data_for_step('1')['output_type']

            # Fetch the query selected in the first step
            query = self.get_cleaned_data_for_step('0')['query']

            return {'output_type': output_type, 'query': query}

        return {}

    def done(self, form_list, **kwargs):
        form_data = [form.cleaned_data for form in form_list]
        
        # Extract the form data
        name = form_data[0]['name']
        description = form_data[0]['description']
        query = form_data[0]['query']
        output_type = form_data[1]['output_type']
        options = {key: value for form in form_data[2:] for key, value in form.items()}

        if options.get('group_by') == '':
            del options['group_by'] 

        # Create Widget instance
        widget = Widget(
            name=name,
            description=description,
            query=query,
            output_type=output_type,
            options=options,
            created_by=self.request.user,
            updated_by=self.request.user
        )

        widget.save()

        return redirect(widget.get_absolute_url())  # Redirect to a new URL


@router.bloomerp_route(
    models=Widget,
    path='preview',
    name='Preview Widget',
    description='Preview a widget',
    route_type='detail',
    url_name='preview'
)
class WidgetPreviewView(BloomerpModelContextMixin, HtmxMixin, DetailView):
    model = Widget
    template_name = 'components/workspace_widget.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['widget'] = self.object
        return context
    
    def get(self, request, *args, **kwargs):
        # Call the components_workspace_widget view with the widget_id GET parameter
        request.GET = request.GET.copy()
        request.GET['widget_id'] = kwargs.get('pk')
        return workspace_widget(request)


@router.bloomerp_route(
    models=Widget,
    path='update/',
    name='Update',
    description='Update widget',
    route_type='detail',
    url_name='update'
)
class WidgetUpdateWizardView(PermissionRequiredMixin, SessionWizardView, BloomerpBaseDetailView):
    form_list = [WidgetForm1, WidgetForm2, WidgetForm3]
    template_name = 'detail_views/bloomerp_detail_wizard_update_view.html'
    permission_required = 'shared_utils.change_widget'
    model = Widget  

    
    def get_context_data(self, **kwargs) -> dict:
        context = super().get_context_data(**kwargs)
        context["exclude_header"] = self.exclude_header
        self.object = self.get_object()
        # This is necessary to get the context data from the BloomerpModelContextMixin
        # Don't ask me why, I don't know
        context.update(BloomerpModelContextMixin.get_context_data(self, **kwargs))

        return context

    def get_form_kwargs(self, step=None):
        """
        Returns the keyword arguments for instantiating the form
        (or formset) on the given step.
        """
        if step == '0':
            return {'widget': self.get_object()}

        if step == '1':
            return {'widget': self.get_object()}

        if step == '2':
            # Fetch the output type selected in the previous step
            output_type = self.get_cleaned_data_for_step('1')['output_type']

            # Fetch the query selected in the first step
            query = self.get_cleaned_data_for_step('0')['query']

            return {'output_type': output_type, 'query': query, 'widget': self.get_object()}

        return {}

    def done(self, form_list, **kwargs):
        form_data = [form.cleaned_data for form in form_list]
        
        # Extract the form data
        name = form_data[0]['name']
        description = form_data[0]['description']
        query = form_data[0]['query']
        output_type = form_data[1]['output_type']
        options = {key: value for form in form_data[2:] for key, value in form.items()}

        if options.get('group_by') == '':
            del options['group_by'] 

        # Get the object
        widget : Widget = self.get_object()

        widget.name = name
        widget.description = description
        widget.query = query
        widget.output_type = output_type
        widget.options = options
        widget.updated_by = self.request.user
        widget.created_by = self.request.user
        
        widget.save()
        return redirect(widget.get_absolute_url())

